__all__ = ('UserBase', )

from re import I as re_ignore_case, compile as re_compile, escape as re_escape

from scarletio import include

from ...bases import DiscordEntity, ICON_TYPE_NONE, Icon, IconSlot, IconType, PlaceHolder, PlaceHolderFunctional
from ...color import Color
from ...http import urls as module_urls
from ...localization.utils import LOCALE_DEFAULT
from ...utils import DATETIME_FORMAT_CODE

from .fields import (
    parse_name, put_avatar_decoration, put_banner_color, put_bot, put_clan, put_discriminator,
    put_display_name, put_flags, put_id, put_name, validate_name
)
from .flags import UserFlag
from .preinstanced import DefaultAvatar, PremiumType, Status


ZEROUSER = include('ZEROUSER')


USER_AVATAR = IconSlot('avatar', 'avatar', module_urls.user_avatar_url, module_urls.user_avatar_url_as)
USER_BANNER = IconSlot('banner', 'banner', module_urls.user_banner_url, module_urls.user_banner_url_as)


class UserBase(DiscordEntity, immortal = True):
    """
    Base class for user instances.
    
    Attributes
    ----------
    avatar_hash : `int`
        The user's avatar's hash in `uint128`.
    avatar_type : ``IconType``
        The user's avatar's type.
    id : `int`
        The client's unique identifier number.
    name : str
        The client's username.
    
    Notes
    -----
    Instances of this type are weakreferable.
    """
    __slots__ = ('name', )
    
    avatar = USER_AVATAR
    
    def __new__(
        cls,
        *,
        avatar = ...,
        name = ...,
    ):
        """
        Creates a new partial user with the given fields.
        
        Parameters
        ----------
        avatar : `None`, ``Icon``, `str`, `bytes-like`, Optional (Keyword only)
            The user's avatar.
        name : `str`, Optional (Keyword only)
            The user's name.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # avatar
        if avatar is ...:
            avatar = None
        else:
            avatar = cls.avatar.validate_icon(avatar, allow_data = True)
        
        # name
        if name is ...:
            name = ''
        else:
            name = validate_name(name)
        
        # Construct
        new = object.__new__(cls)
        new.avatar = avatar
        new.id = 0
        new.name = name
        return new
    
    
    def _update_attributes(self, data):
        """
        Updates the user with the given data by overwriting it's old attributes.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            User data received from Discord.
        """
        self._set_avatar(data)
        self.name = parse_name(data)
    
    
    def _difference_update_attributes(self, data):
        """
        Updates the user and returns it's overwritten attributes as a `dict` with a `attribute-name` - `old-value`
        relation.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            User data received from Discord.
        
        Returns
        -------
        old_attributes : `dict` of (`str`, `object`) items
            All item in the returned dict is optional.
        
        Returned Data Structure
        -----------------------
        
        +-----------------------+-------------------------------+-----------------------+
        | Keys                  | Values                        | Applicable for        |
        +=======================+===============================+=======================+
        | avatar                | ``Icon``                      | all                   |
        +-----------------------+-------------------------------+-----------------------+
        | avatar_decoration     | `None`, ``AvatarDecoration``  | ``Client``, ``User``  |
        +-----------------------+-------------------------------+-----------------------+
        | banner                | ``Icon``                      | ``Client``, ``User``  |
        +-----------------------+-------------------------------+-----------------------+
        | banner_color          | `None`, ``Color``             | ``Client``, ``User``  |
        +-----------------------+-------------------------------+-----------------------+
        | channel_id            | `int`                         | ``Webhook``           |
        +-----------------------+-------------------------------+-----------------------+
        | clan                  | `None`, ``UserClan``          | ``Client``, ``User``  |
        +-----------------------+-------------------------------+-----------------------+
        | discriminator         | `int`                         | ``Client``, ``User``  |
        +-----------------------+-------------------------------+-----------------------+
        | display_name          | `None`, `str`                 | ``Client``, ``User``  |
        +-----------------------+-------------------------------+-----------------------+
        | email                 | `None`, `str`                 | ``Client``            |
        +-----------------------+-------------------------------+-----------------------+
        | email_verified        | `bool`                        | ``Client``            |
        +-----------------------+-------------------------------+-----------------------+
        | flags                 | ``UserFlag``                  | ``Client``, ``User``  |
        +-----------------------+-------------------------------+-----------------------+
        | locale                | ``Locale``                    | ``Client``            |
        +-----------------------+-------------------------------+-----------------------+
        | mfa_enabled           | `bool`                        | ``Client``            |
        +-----------------------+-------------------------------+-----------------------+
        | name                  | `str`                         | all                   |
        +-----------------------+-------------------------------+-----------------------+
        | premium_type          | ``PremiumType``               | ``Client``            |
        +-----------------------+-------------------------------+-----------------------+
        """
        old_attributes = {}
        
        self._update_avatar(data, old_attributes)
        
        name = parse_name(data)
        if self.name != name:
            old_attributes['name'] = self.name
            self.name = name
        
        return old_attributes
    
    
    @classmethod
    def _create_empty(cls, user_id):
        """
        Creates a user with its default attributes values set.
        
        Parameters
        ----------
        user_id : `int`
            The user's identifier.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.id = user_id
        self._set_default_attributes()
        return self
    
    
    def _set_default_attributes(self):
        """
        Sets the user's attribute's to their default.
        """
        self.avatar_hash = 0
        self.avatar_type = ICON_TYPE_NONE
        self.name = ''
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new user from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            User data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        raise NotImplementedError(
            f'`{cls.__name__}` does not support `.from_data` operation, please call it on a sub-type of it.'
        )
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Tries to convert the user back to a json serializable dictionary.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        data = {}
        
        USER_AVATAR.put_into(self.avatar, data, defaults, as_data = not include_internals)
        USER_BANNER.put_into(self.banner, data, defaults, as_data = not include_internals)
        put_avatar_decoration(self.avatar_decoration, data, defaults)
        put_banner_color(self.banner_color, data, defaults)
        put_discriminator(self.discriminator, data, defaults)
        put_display_name(self.display_name, data, defaults)
        put_name(self.name, data, defaults)
        
        if include_internals:
            put_bot(self.bot, data, defaults)
            put_clan(self.clan, data, defaults)
            put_id(self.id, data, defaults)
            put_flags(self.flags, data, defaults)
        
        return data
    
    
    def copy(self):
        """
        Copies the user returning a new partial one.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.avatar = self.avatar
        new.id = 0
        new.name = self.name
        return new
    
    
    def copy_with(
        self,
        *,
        avatar = ...,
        name = ...,
    ):
        """
        Copies the user with the given fields.
        
        Parameters
        ----------
        avatar : `None`, ``Icon``, `str`, `bytes-like`, Optional (Keyword only)
            The user's avatar.
        name : `str`, Optional (Keyword only)
            The user's name.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # avatar
        if avatar is ...:
            avatar = self.avatar
        else:
            avatar = type(self).avatar.validate_icon(avatar, allow_data = True)
        
        # name
        if name is ...:
            name = self.name
        else:
            name = validate_name(name)
        
        # Construct
        new = object.__new__(type(self))
        new.avatar = avatar
        new.id = 0
        new.name = name
        return new
    
    
    def __repr__(self):
        """Returns the user's representation."""
        repr_parts = [
            '<',
            type(self).__name__,
        ]
        
        user_id = self.id
        if user_id:
            repr_parts.append(' id = ')
            repr_parts.append(repr(user_id))
            repr_parts.append(',')
        
        repr_parts.append(' name = ')
        repr_parts.append(repr(self.full_name))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __format__(self, code):
        """
        Formats the user in a format string.
        
        Parameters
        ----------
        code : `str`
            The option on based the result will be formatted.
        
        Returns
        -------
        user : `str`
        
        Raises
        ------
        ValueError
            Unknown format code.
        
        Examples
        --------
        ```py
        >>> from hata import User, now_as_id
        >>> user = User.precreate(now_as_id(), name = 'Neko', discriminator = 2012)
        >>> user
        <User id = 730233383967260672, name = 'Neko#2012'>
        >>> # no code stands for `user.name`.
        >>> f'{user}'
        'Neko'
        >>> # 'f' stands for full name
        >>> f'{user:f}'
        'Neko#2012'
        >>> # 'm' stands for mention.
        >>> f'{user:m}'
        '<@730233383967260672>'
        >>> # 'c' stands for created at.
        >>> f'{user:c}'
        '2020.07.08-01:26:45'
        ```
        """
        if not code:
            return self.name
        
        if code == 'f':
            return self.full_name
        
        if code == 'm':
            return self.mention
        
        if code == 'c':
            return format(self.created_at, DATETIME_FORMAT_CODE)
        
        raise ValueError(
            f'Unknown format code {code!r} for {self.__class__.__name__}; {self!r}. '
            f'Available format codes: {""!r}, {"c"!r}, {"f"!r}, {"m"!r}.'
        )
    
    def __hash__(self):
        """Returns the user's hash."""
        user_id = self.id
        if user_id or (self is ZEROUSER):
            return user_id
        
        return self._get_hash_partial()
    
    
    def _get_hash_partial(self):
        """
        Returns a partial user's hash value.
        
        Returns
        -------
        hash_value : `int`
        """
        hash_value = 0
        
        # avatar
        hash_value ^= hash(self.avatar)
        
        # id | Internal -> skip
        
        # name
        hash_value ^= hash(self.name)
        
        return hash_value
    
    
    # Sorting
    
    def __gt__(self, other):
        """Returns whether the user's id is greater than the other's."""
        if not isinstance(other, UserBase):
            return NotImplemented
        
        return self.id > other.id
    
    
    def __ge__(self, other):
        """Returns whether the user's id is greater or equal to the other."""
        if not isinstance(other, UserBase):
            return NotImplemented
        
        if self.id > other.id:
            return True
        
        return self._is_equal_same_type(other)
    
    
    def __eq__(self, other):
        """Return whether the user's id is equal to the other."""
        if not isinstance(other, UserBase):
            return NotImplemented
        
        return self._is_equal_same_type(other)
    
    
    def __ne__(self, other):
        """Returns whether the user's id is different as the other's."""
        if not isinstance(other, UserBase):
            return NotImplemented
        
        return not self._is_equal_same_type(other)
    
    
    def __le__(self, other):
        """Returns whether the user's id is less or equal to the other."""
        if not isinstance(other, UserBase):
            return NotImplemented
        
        if self.id < other.id:
            return True
        
        return self._is_equal_same_type(other)
    
    
    def __lt__(self, other):
        """Returns whether the user's id is less than the other's."""
        if not isinstance(other, UserBase):
            return NotImplemented
        
        return self.id < other.id
    
    
    def _is_equal_same_type(self, other):
        """
        Returns whether the two users are equal. `self` and `other` must be the same type.
        
        Parameters
        ----------
        other : `instance<type<self>>`
            The other user.
        
        Returns
        -------
        is_equal : `bool`
        """
        self_id = self.id
        other_id = other.id
        if ((self_id or self is ZEROUSER) and (other_id or other is ZEROUSER)):
            return self_id == other_id
        
        return self._compare_attributes(other)
    
    
    def _compare_user_attributes_extended(self, other):
        """
        Compares the two user's user attributes (excluding id obviously).
        
        Parameters
        ----------
        other : `instance<type<self>>`
            The other user.
        
        Returns
        -------
        is_equal : `bool`
        """
        # activities
        if (self.activities != other.activities):
            return False
        
        # avatar_hash
        if (self.avatar_hash != other.avatar_hash):
            return False
        
        # avatar_type
        if (self.avatar_type != other.avatar_type):
            return False
        
        # avatar_decoration
        if (self.avatar_decoration != other.avatar_decoration):
            return False
        
        # banner_color
        if (self.banner_color != other.banner_color):
            return False
        
        # banner_hash
        if (self.banner_hash != other.banner_hash):
            return False
        
        # banner_type
        if (self.banner_type != other.banner_type):
            return False
        
        # bot
        if (self.bot != other.bot):
            return False
        
        # clan
        if (self.clan != other.clan):
            return False
        
        # discriminator
        if (self.discriminator != other.discriminator):
            return False
        
        # display_name
        if (self.display_name != other.display_name):
            return False
        
        # email
        if (self.email != other.email):
            return False
        
        # email_verified
        if (self.email_verified != other.email_verified):
            return False
        
        # flags
        if (self.flags != other.flags):
            return False
        
        # locale
        if (self.locale is not other.locale):
            return False
        
        # mfa_enabled
        if (self.mfa_enabled != other.mfa_enabled):
            return False
        
        # name
        if (self.name != other.name):
            return False
        
        # premium_type
        if (self.premium_type is not other.premium_type):
            return False
        
        # status
        if (self.status is not other.status):
            return False
        
        # statuses
        if (self.statuses != other.statuses):
            return False
        
        return True
    
    
    _compare_attributes = _compare_user_attributes_extended
    
    # Place holders
    
    activities = PlaceHolder(
        None,
        """
        Returns the user's activities.
        
        Returns
        -------
        activities : `None`, `list` of ``Activity``
        """
    )
    
    
    avatar_decoration = PlaceHolder(
        None,
        """
        Returns the user's avatar decoration.
        
        Returns
        -------
        avatar_decoration : `None`, ``AvatarDecoration``
        """
    )
    
    
    banner = PlaceHolderFunctional(
        (lambda : Icon(IconType.none, 0)),
        """
        Returns the user's banner.
        
        Returns
        -------
        banner : ``Icon``
        """
    )
    
    
    banner_color = PlaceHolder(
        None,
        """
        Returns the user's banner color.
        
        Returns
        -------
        banner_color : `None`, ``Color``
        """
    )
    
    
    banner_hash = PlaceHolder(
        0,
        """
        Returns the user's banner's hash.
        
        Returns
        -------
        banner_hash : `int`
        """,
    )
    
    
    banner_type = PlaceHolder(
        IconType.none,
        """
        Returns the user's banner's type.
        
        Returns
        -------
        banner_type : ``IconType``
        """,
    )
    
    
    avatar_decoration_url = property(module_urls.user_avatar_decoration_url)
    avatar_decoration_url_as = module_urls.user_avatar_decoration_url_as
    
    
    bot = PlaceHolder(
        False,
        """
        Returns whether the user is a bot or a user account.
        
        Returns
        -------
        bot : `bool`
        """,
    )
    
    clan = PlaceHolder(
        None,
        """
        Returns the user's primary clan.
        
        Returns
        -------
        clan : `None`, ``UserClan``
        """,
    )
    
    discriminator = PlaceHolder(
        0,
        """
        Returns the user's discriminator.
        
        Returns
        -------
        discriminator : `bool`
        """,
    )
    
    
    display_name = PlaceHolder(
        None,
        """
        Returns the user's non-unique display name.
        
        Returns
        -------
        display_name : `None`, `str`
        """,
    )
    
    
    email = PlaceHolder(
        None,
        """
        Returns the user's email.
        
        Returns
        -------
        email : `None`, `str`
        """,
    )
    
    
    email_verified = PlaceHolder(
        False,
        """
        Returns the user's email is verified.
        
        Returns
        -------
        email_verified : `bool`
        """,
    )
    
    
    flags = PlaceHolder(
        UserFlag(),
        """
        Returns the user's flags.
        
        Returns
        -------
        flags : ``UserFlag``
        """
    )
    
    
    guild_profiles = PlaceHolderFunctional(
        (lambda : {}),
        """
        Returns a dictionary, which contains the user's guild profiles. If the user is member of a guild, then it should
        have a respective guild profile accordingly.
        
        Returns
        -------
        guild_profiles : `dict` of (`int`, ``GuildProfile``) items
        """
    )
    
    
    locale = PlaceHolder(
        LOCALE_DEFAULT,
        """
        Returns preferred locale by the user.
        
        Returns
        -------
        locale : ``Locale``
        """
    )
    
    
    mfa_enabled = PlaceHolder(
        False,
        """
        Returns whether the user has two factor authorization enabled on the account.
        
        Returns
        -------
        mfa_enabled : `bool`
        """
    )
    
    
    premium_type = PlaceHolder(
        PremiumType.none,
        """
        Returns the Nitro subscription type of the user.
        
        Returns
        -------
        premium_type : ``PremiumType``
        """
    )
    
    
    status = PlaceHolder(
        Status.offline,
        """
        Returns the user's display status.
        
        Returns
        -------
        status  ``Status``
        """
    )
    
    
    statuses = PlaceHolder(
        None,
        """
        Returns the user's statuses for each platform.
        
        Returns
        -------
        statuses : `None`, `dict` of (`str`, `str`) items
        """
    )
    
    
    thread_profiles = PlaceHolder(
        None,
        """
        Returns the user's activities.
        
        Returns
        -------
        activities : `None`, `tuple` of ``Activity``
        """
    )
    
    # Properties
    
    @property
    def full_name(self):
        """
        The user's name with it's discriminator.
        
        Returns
        -------
        full_name : `str`
        """
        discriminator = self.discriminator
        if discriminator:
            return f'{self.name}#{discriminator:0>4}'
        
        return self.name
    
    
    @property
    def mention(self):
        """
        The mention of the user.
        
        Returns
        -------
        mention : `str`
        """
        return f'<@{self.id}>'
    
    
    @property
    def mention_nick(self):
        """
        The mention to the user's nick.
        
        Returns
        -------
        mention : `str`
        
        Notes
        -----
        It actually has nothing to do with the user's nickname > <.
        """
        return f'<@!{self.id}>'
    
    
    @property
    def default_avatar_url(self):
        """
        Returns the user's default avatar's url.
        
        Returns
        -------
        default_avatar_url : `str`
        """
        return DefaultAvatar.for_(self).url
    
    
    @property
    def default_avatar(self):
        """
        Returns the user's default avatar.
        
        Returns
        -------
        default_avatar : ``DefaultAvatar``
        """
        return DefaultAvatar.for_(self)
    
    
    def iter_activities(self):
        """
        Iterates over the user's activities.
        
        This method is an iterable generator.
        
        Yields
        ------
        activity : ``Activity``
        """
        activities = self.activities
        if (activities is not None):
            yield from activities
    
    
    @property
    def partial(self):
        """
        Returns whether the user is partial. Partial users have only their ``.id`` set and every other field might not
        reflect the reality.
        
        Returns
        -------
        partial : `bool`
        """
        return True
    
    
    @property
    def activity(self):
        """
        Returns the user's top activity if applicable. If not.
        
        Returns
        -------
        activity : ``Activity``, `None`
        """
        return None
    
    
    @property
    def custom_activity(self):
        """
        Returns the user's custom activity if applicable.
        
        Returns
        -------
        activity : ``Activity``, `None`
        """
        return None
    
    
    @property
    def platform(self):
        """
        Returns the user's top status's platform. If the user is offline it will return an empty string.
        
        Returns
        -------
        platform : `str`
        """
        return ''
    
    
    def get_status_by_platform(self, platform):
        """
        Gets the user's status by the given platform.
        
        Parameters
        ----------
        platform : `str`
            The platform to get the status for.
        
        Returns
        -------
        status : ``Status``
            Defaults to `status.offline` if the user has no specific status for the given platform.
        """
        return Status.offline
    
    
    def color_at(self, guild):
        """
        Returns the user's color at the given guild.
        
        Parameters
        ----------
        guild : `None`, ``Guild``, `int`
            The guild, where the user's color will be checked.
            
            Can be given as `None`.

        Returns
        -------
        color : ``Color``
        """
        return Color()
    
    
    def name_at(self, guild):
        """
        Returns the user's name at the given guild.
        
        Parameters
        ----------
        guild : `None`, ``Guild``, `int`
            The guild, where the user's nick will be checked.
            
            Can be given as `None`.

        Returns
        -------
        name : `str`
        """
        display_name = self.display_name
        if (display_name is not None):
            return display_name
        
        return self.name
    
    
    def has_name_like(self, name):
        """
        Returns whether the user's name is like the given string.
        
        Parameters
        ----------
        name : `str`
            The name of the user.
        
        Returns
        -------
        has_name_like : `bool`
        """
        if name.startswith('@'):
            name = name[1:]
        
        name_length = len(name)
        if name_length < 1:
            return False
        
        if name_length > 5:
            if name_length > 37:
                return False
            
            if name[-5] == '#':
                try:
                    discriminator = int(name[-4:])
                except ValueError:
                    pass
                else:
                    stripped_name = name[:-5]
                    if (self.discriminator == discriminator) and (self.name == stripped_name):
                        return True
        
        if name_length > 32:
            return False
        
        pattern = re_compile(re_escape(name), re_ignore_case)
        
        if pattern.search(self.name) is not None:
            return True
        
        display_name = self.display_name
        if (display_name is not None) and (pattern.search(display_name) is not None):
            return True
        
        return False
    
    
    def has_name_like_at(self, name, guild):
        """
        Returns whether the user's name is like the given string.
        
        Parameters
        ----------
        name : `str`
            The name of the user.
        
        guild : `None`, ``Guild``, `int`
            The guild, where the user's nick will be also checked.
            
            Can be given as `None`.
        
        Returns
        -------
        has_name_like : `bool`
        """
        return self.has_name_like(name)
    
    
    def mentioned_in(self, message):
        """
        Returns whether the user is mentioned at a given message.
        
        Parameters
        ----------
        message : ``Message``
            The message, what's mentions will be checked.
        
        Returns
        -------
        mentioned : `bool`
        """
        if message.mentioned_everyone:
            return True
        
        mentioned_users = message.mentioned_users
        if (mentioned_users is not None) and (self in mentioned_users):
            return True
        
        mentioned_roles = message.mentioned_roles
        if (mentioned_roles is not None):
            guild_id = message.guild_id
            if guild_id:
                try:
                    guild_profile = self.guild_profiles[guild_id]
                except KeyError:
                    pass
                else:
                    roles = guild_profile.roles
                    if (roles is not None):
                        for role in roles:
                            if role in mentioned_roles:
                                return True
        
        return False
    
    
    def has_role(self, role):
        """
        Returns whether the user has the given role.
        
        Parameters
        ----------
        role : ``Role``
            The role what will be checked.

        Returns
        -------
        has_role : `bool`
        """
        return False
    
    
    def top_role_at(self, guild, default = None):
        """
        Returns the top role of the user at the given guild.
        
        Parameters
        ----------
        guild : `None`, ``Guild``, `int`
            The guild where the user's top role will be looked up.
        
        default : `object` = `None`, Optional
            If the user is not a member of the guild, or if has no roles there, then the given default value is returned.
            Defaults to `None`.
        
        Returns
        -------
        top_role ``Role``, `default`
        """
        return default
    
    
    def can_use_emoji(self, emoji):
        """
        Returns whether the user can use the given emoji.
        
        Parameters
        ----------
        emoji : ``Emoji``
            The emoji to check.
        
        Returns
        -------
        can_use_emoji : `bool`
        """
        if emoji.is_unicode_emoji():
            return True
        
        return False
    
    
    def has_higher_role_than(self, role):
        """
        Returns whether the user has higher role than the given role at it's respective guild.
        
        If the user is the owner of the guild, then returns `True` even if it has lower role.
        
        Parameters
        ----------
        role : ``Role``
            The role to check.
        
        Returns
        -------
        has_higher_role_than : `bool`
        """
        return False
    
    
    def has_higher_role_than_at(self, user, guild):
        """
        Returns whether the user has higher role as the other one at the given guild.
        
        Parameters
        ----------
        user : ``User``
            The other user to check.
        guild : ``Guild``, `None`, `int`
            The guild where the users' top roles will be checked.
            
            Can be given as `None`.
        
        Returns
        -------
        has_higher_role_than_at : `bool`
        """
        return False
    
    
    def get_guild_profile_for(self, guild):
        """
        Returns the user's guild profile for the given guild.
        
        Parameters
        ----------
        guild : `None`, ``Guild``, `int`
            The guild to get guild profile for.
        
        Returns
        -------
        guild_profile : `None`, ``Guild``
        """
        return None
    
    
    def iter_guilds_and_profiles(self):
        """
        Iterates over the guild profiles of the user.
        
        This method is an iterable generator.
        
        Yields
        ------
        guild : ``Guild``
            The guild profile's guild.
        guild_profile : ``GuildProfile``
            The user's guild profile in the guild.
        """
        return
        yield
    
    
    def iter_guilds(self):
        """
        Iterates over the guilds of the user.
        
        This method is an iterable generator.
        
        Yields
        ------
        guild : ``Guild``
            The guild profile's guild.
        """
        return
        yield
    
    
    def is_boosting(self, guild):
        """
        Returns whether the user is boosting the given guild.
        
        Parameters
        ----------
        guild : `None`, ``Guild``, `int`
            The guild to get whether the user is booster of.
        
        Returns
        -------
        is_boosting : `bool`
        """
        return False
    
    
    # User base does not have `.banner` assigned, so manually add banner functions
    banner_url = property(module_urls.user_banner_url)
    banner_url_as = module_urls.user_banner_url_as
    
    avatar_url_for = module_urls.user_avatar_url_for
    avatar_url_for_as = module_urls.user_avatar_url_for_as
    avatar_url_at = module_urls.user_avatar_url_at
    avatar_url_at_as = module_urls.user_avatar_url_at_as
    
    banner_url_for = module_urls.user_banner_url_for
    banner_url_for_as = module_urls.user_banner_url_for_as
    banner_url_at = module_urls.user_banner_url_at
    banner_url_at_as = module_urls.user_banner_url_at_as
    
    
    def _delete(self):
        """
        Deletes the user from it's guilds.
        """
        pass
    
    
    def _difference_update_presence(self, data):
        """
        Updates the user's presence and returns it's overwritten attributes as a `dict` with a `attribute-name` -
        `old-value` relation. An exception from this is `activities`, because that's a ``ActivityChange``
        containing all the changes of the user's activities.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Received guild member data.
        
        Returns
        -------
        old_attributes : `dict` of (`str`, `object`) items
            All item in the returned dict is optional.
        
        Returned Data Structure
        -----------------------
        
        +---------------+-------------------------------------------+
        | Keys          | Values                                    |
        +===============+===========================================+
        | activities    | ``ActivityChange``                        |
        +---------------+-------------------------------------------+
        | status        | ``Status``                                |
        +---------------+-------------------------------------------+
        | statuses      | `None`, `dict` of (`str`, `str`) items    |
        +---------------+-------------------------------------------+
        """
        return {}
    
    
    def _update_presence(self, data):
        """
        Updates the user's presences with the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Received guild member data.
        """
        pass
    
    
    @classmethod
    def _from_data_and_update_profile(cls, data, guild):
        """
        First tries to find the user, then it's respective guild profile for the given guild to update it.
        
        If the method cannot find the user, or the respective guild profile, then creates them.
        
        Not like ``._from_data_and_difference_update_profile``, this method not calculates changes.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Received guild member data.
        guild : ``Guild``
            The respective guild of the profile to update.
        
        Returns
        -------
        user : ``ClientUserbase``
            The updated user.
        """
        raise NotImplementedError
    
    
    def _update_profile(self, data, guild):
        """
        Tries to find the user's respective guild profile. If it cannot find it, creates it.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Received guild profile data.
        guild : ``Guild``
            The respective guild of the profile to update.
        
        Returns
        -------
        updated : `bool`
            Returns `True` if the guild profile was updated and `False` when created.
        """
        raise NotImplementedError
    
    
    @classmethod
    def _from_data_and_difference_update_profile(cls, data, guild):
        """
        First tries to find the user, then it's respective guild profile for the given guild to update it.
        
        If the method cannot find the user, or the respective guild profile, then creates them.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Received guild member data.
        guild : ``Guild``
            The respective guild of the profile to update.

        Returns
        -------
        user : ``ClientUserBase``
            The respective user.
        old_attributes : `None | dict<str, object>`
            The changed attributes of the respective guild profile as a `dict` with `attribute-name` - `old-attribute`
            relation.
            
            May return `None` indicating that the guild profile was not cached before so we cannot determine whether
            it would have been updated.
            
            The possible keys and values within `old_attributes` are all optional and they can be any of the following:
            +-------------------+-------------------------------+
            | Keys              | Values                        |
            +===================+===============================+
            | avatar            | ``Icon``                      |
            +-------------------+-------------------------------+
            | avatar_decoration | `None`, ``AvatarDecoration``  |
            +-------------------+-------------------------------+
            | banner            | ``Icon``                      |
            +-------------------+-------------------------------+
            | boosts_since      | `None`, `DateTime`            |
            +-------------------+-------------------------------+
            | flags             | `None`, ``GuildProfileFlags`` |
            +-------------------+-------------------------------+
            | nick              | `None`, `str`                 |
            +-------------------+-------------------------------+
            | pending           | `bool`                        |
            +-------------------+-------------------------------+
            | role_ids          | `None`, `tuple` of `int`      |
            +-------------------+-------------------------------+
            | timed_out_until   | `None`, `DateTime`            |
            +-------------------+-------------------------------+
        """
        raise NotImplementedError
    
    
    def _difference_update_profile(self, data, guild):
        """
        Tries to find the user's respective guild profile. If it cannot find it, creates it.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Received guild profile data.
        guild : ``Guild``
            The respective guild of the profile to update.
        
        Returns
        -------
        old_attributes : `None | dict<str, object>`
            The changed attributes of the respective guild profile as a `dict` with `attribute-name` - `old-attribute`
            relation.
            
            May return `None` indicating that the guild profile was not cached before so we cannot determine whether
            it would have been updated.
            
            The possible keys and values within `old_attributes` are all optional and they can be any of the following:
            +-------------------+-------------------------------+
            | Keys              | Values                        |
            +===================+===============================+
            | avatar            | ``Icon``                      |
            +-------------------+-------------------------------+
            | avatar_decoration | `None`, ``AvatarDecoration``  |
            +-------------------+-------------------------------+
            | banner            | ``Icon``                      |
            +-------------------+-------------------------------+
            | boosts_since      | `None`, `DateTime`            |
            +-------------------+-------------------------------+
            | flags             | `None`, ``GuildProfileFlags`` |
            +-------------------+-------------------------------+
            | nick              | `None`, `str`                 |
            +-------------------+-------------------------------+
            | pending           | `bool`                        |
            +-------------------+-------------------------------+
            | role_ids          | `None`, `tuple` of `int`      |
            +-------------------+-------------------------------+
            | timed_out_until   | `None`, `DateTime`            |
            +-------------------+-------------------------------+
        """
        raise NotImplementedError



__all__ = ('AuditLogEntryChangeConversion', )

from scarletio import RichAttributeErrorBaseType

from ..conversion_helpers.helpers import _eq_functions, _hash_function

from .change_deserializers import change_deserializer_modification
from .change_serializers import change_serializer_modification


class AuditLogEntryChangeConversion(RichAttributeErrorBaseType):
    """
    Represents how changes are converted of an audit log entry.
    
    Attributes
    ----------
    change_deserialization_key_pre_check : `None | FunctionType | MethodType`
        In case the deserialization is not key matched, this can be used to match the keys instead.
    change_deserializer : `GeneratorFunctionType`
        Change deserializers to use.
    change_serializer : `GeneratorFunctionType`
        Change serialiser to use.
    field_keys : `None | tuple<str>`
        The serialised name of the field.
    field_name : `str`
        The field's name.
    value_deserializer : `None | FunctionType | MethodType`
        Raw to processed value converter.
    value_merger : `None | FunctionType | MethodType`
        Merger to use to merge two entries.
    value_serializer : `None | FunctionType | MethodType`
        Processed to raw value converter.
    value_validator : `FunctionType | MethodType`
        Change value validator.
    """
    __slots__ = (
        'change_deserialization_key_pre_check', 'change_deserializer', 'change_serializer', 'field_keys', 'field_name',
        'value_deserializer', 'value_merger', 'value_serializer', 'value_validator'
    )
    
    def __new__(
        cls,
        field_keys,
        field_name,
        *,
        change_deserialization_key_pre_check = ...,
        change_deserializer = ...,
        change_serializer = ...,
        value_deserializer = ...,
        value_merger = ...,
        value_serializer = ...,
        value_validator = ...
    ):
        """
        Creates a new change conversion representation with the given fields.
        
        Parameters
        ----------
        field_keys : `None | tuple<str>`
            The serialised name of the field.
        field_name : `str`
            The field's name.
        change_deserializer  : `GeneratorFunctionType`, Optional (Keyword only)
            Deserializer to use.
        change_deserialization_key_pre_check : `FunctionType | MethodType`, Optional (Keyword only)
            In case the deserialization is not key matched, this can be used to match the keys instead.
        change_serializer : `GeneratorFunctionType`, Optional (Keyword only)
            Change serialiser to use.
        value_deserializer : `FunctionType | MethodType`, Optional (Keyword only)
            Raw to processed value converter.
        value_merger : `FunctionType | MethodType`, Optional (Keyword only)
            Merger to use to merge two entries.
        value_serializer : `FunctionType | MethodType`, Optional (Keyword only)
            Processed to raw value converter.
        value_validator : `FunctionType | MethodType`, Optional (Keyword only)
            Change value validator.
        """
        if change_deserialization_key_pre_check is ...:
            change_deserialization_key_pre_check = None
        
        if change_deserializer is ...:
            change_deserializer = change_deserializer_modification
        
        if change_serializer is ...:
            change_serializer = change_serializer_modification
        
        if value_deserializer is ...:
            value_deserializer = None
        
        if value_merger is ...:
            value_merger = None
        
        if value_serializer is ...:
            value_serializer = None
        
        if value_validator is ...:
            value_validator = None
        
        self = object.__new__(cls)
        self.change_deserializer = change_deserializer
        self.change_deserialization_key_pre_check = change_deserialization_key_pre_check
        self.change_serializer = change_serializer
        self.value_merger = value_merger
        self.field_keys = field_keys
        self.field_name = field_name
        self.value_deserializer = value_deserializer
        self.value_serializer = value_serializer
        self.value_validator = value_validator
        return self
    
    
    def __repr__(self):
        """Returns the change conversion's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        # field_keys
        repr_parts.append(' field_keys = ')
        repr_parts.append(repr(self.field_keys))
        
        # field_name
        repr_parts.append(', field_name = ')
        repr_parts.append(repr(self.field_name))
        
        # change_deserializer
        change_deserializer = self.change_deserializer
        repr_parts.append(', change_deserializer = ')
        repr_parts.append(change_deserializer.__module__)
        repr_parts.append('.')
        repr_parts.append(change_deserializer.__name__)
        
        # change_serializer
        change_serializer = self.change_serializer
        repr_parts.append(', change_serializer = ')
        repr_parts.append(change_serializer.__module__)
        repr_parts.append('.')
        repr_parts.append(change_serializer.__name__)
        
        # change_deserialization_key_pre_check
        change_deserialization_key_pre_check = self.change_deserialization_key_pre_check
        change_deserialization_key_pre_check = self.value_merger
        if (change_deserialization_key_pre_check is not None):
            repr_parts.append(', change_deserialization_key_pre_check = ')
            repr_parts.append(change_deserialization_key_pre_check.__module__)
            repr_parts.append('.')
            repr_parts.append(change_deserialization_key_pre_check.__name__)
        
        # value_merger
        value_merger = self.value_merger
        if (value_merger is not None):
            repr_parts.append(', value_merger = ')
            repr_parts.append(value_merger.__module__)
            repr_parts.append('.')
            repr_parts.append(value_merger.__name__)
        
        
        # value_deserializer
        value_deserializer = self.value_deserializer
        if (value_deserializer is not None):
            repr_parts.append(', value_deserializer = ')
            repr_parts.append(value_deserializer.__module__)
            repr_parts.append('.')
            repr_parts.append(value_deserializer.__name__)
        
        # value_serializer
        value_serializer = self.value_serializer
        if (value_serializer is not None):
            repr_parts.append(', value_serializer = ')
            repr_parts.append(value_serializer.__module__)
            repr_parts.append('.')
            repr_parts.append(value_serializer.__name__)
        
        # value_validator
        value_validator = self.value_validator
        if (value_validator is not None):
            repr_parts.append(', value_validator = ')
            repr_parts.append(value_validator.__module__)
            repr_parts.append('.')
            repr_parts.append(value_validator.__name__)
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the entry change conversion's hash value."""
        hash_value = 0
        
        # change_deserialization_key_pre_check
        hash_value ^= _hash_function(self.change_deserialization_key_pre_check)
        
        # change_deserializer
        hash_value ^= _hash_function(self.change_deserializer)
        
        # change_serializer
        hash_value ^= _hash_function(self.change_serializer)
        
        # field_keys
        field_keys = self.field_keys
        if (field_keys is not None):
            hash_value ^= hash(field_keys)
        
        # field_name
        hash_value ^= hash(self.field_name)
        
        # value_deserializer
        hash_value ^= _hash_function(self.value_deserializer)
        
        # value_merger
        hash_value ^= hash(self.value_merger)
        
        # value_serializer
        hash_value ^= _hash_function(self.value_serializer)
        
        # value_validator
        hash_value ^= _hash_function(self.value_validator)
        
        return hash_value
    
    
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        # change_deserialization_key_pre_check
        if not _eq_functions(self.change_deserialization_key_pre_check, other.change_deserialization_key_pre_check):
            return False
        
        # change_deserializer
        if not _eq_functions(self.change_deserializer, other.change_deserializer):
            return False
        
        # change_serializer
        if not _eq_functions(self.change_serializer, other.change_serializer):
            return False
        
        # field_keys
        if self.field_keys != other.field_keys:
            return False
        
        # field_name
        if self.field_name != other.field_name:
            return False
        
        # value_deserializer
        if not _eq_functions(self.value_deserializer, other.value_deserializer):
            return False
        
        # value_merger
        if not _eq_functions(self.value_merger, other.value_merger):
            return False
        
        # value_serializer
        if not _eq_functions(self.value_serializer, other.value_serializer):
            return False
        
        # value_validator
        if not _eq_functions(self.value_validator, other.value_validator):
            return False
        
        return True
    
    
    def iter_field_keys(self):
        """
        Iterates over the field keys of the audit log change conversion.
        
        This method is an iterable generator.
        
        Yields
        ------
        field_key : `str`
        """
        field_keys = self.field_keys
        if (field_keys is not None):
            yield from field_keys
    
    
    def get_field_key(self, index):
        """
        Gets the field key at the given index.
        
        Parameters
        ----------
        index : `int`
            The index to get the field key at.
        
        Returns
        -------
        field_key : `None | str`
        """
        field_keys = self.field_keys
        if (field_keys is None):
            return None
        
        if index < 0 or index >= len(field_keys):
            return None
        
        return field_keys[index]
    
    
    def set_value_deserializer(self, value_deserializer):
        """
        Sets the get converter of the conversion and returns it back.
        
        Parameters
        ----------
        value_deserializer : `FunctionType | MethodType`
            Converter to put.
        
        Returns
        -------
        value_deserializer : `value_deserializer`
        """
        self.value_deserializer = value_deserializer
        return value_deserializer
    
    
    def set_value_serializer(self, value_serializer):
        """
        Sets the put converter of the conversion and returns it back.
        
        Parameters
        ----------
        value_serializer : `FunctionType | MethodType`
            Converter to put.
        
        Returns
        -------
        value_serializer : `value_serializer`
        """
        self.value_serializer = value_serializer
        return value_serializer
    
    
    def set_value_validator(self, value_validator):
        """
        Sets the converter of the conversion and returns it back.
        
        Parameters
        ----------
        value_validator : `FunctionType | MethodType`
            Converter to put.
        
        Returns
        -------
        value_validator : `value_validator`
        """
        self.value_validator = value_validator
        return value_validator
    
    
    def should_add_as_key_based_deserializer(self):
        """
        Whether the conversion should be added as a key based deserializer.
        
        Returns
        -------
        should_add_as_key_based_deserializer : `bool`
        """
        if self.field_keys is None:
            return False
        
        if self.change_deserialization_key_pre_check is not None:
            return False
        
        return True
    
    
    def should_add_as_key_pre_check_deserializer(self):
        """
        Whether the conversion should be added as a default deserializer.
        
        Returns
        -------
        should_add_as_key_pre_check_deserializer : `bool`
        """
        if self.change_deserialization_key_pre_check is None:
            return False
        
        return True
    
    
    def should_add_by_field_name(self):
        """
        Whether the conversion should be registered into a group based on its ``-field_name``.
        
        Returns
        -------
        should_add_by_field_name : `str`
        """
        if not self.field_name:
            return False
        
        return True



__all__ = ()


def value_merger_sorted_array(value_0, value_1):
    """
    Merges two sorted array.
    
    Parameters
    ----------
    value_0 : `None | tuple<sortable>`
        Self instance's value.
    value_1 : `None | tuple<sortable>`
        Other instance's value.
    
    Returns
    -------
    new : `None | tuple<sortable>`
    """
    if value_0 is None:
        return value_1
    
    if value_1 is None:
        return value_0
    
    merged = [*value_0, *value_1]
    merged.sort()
    
    return tuple(merged)


def value_merger_replace(value_0, value_1):
    """
    Merges any 2 values.
    
    Parameters
    ----------
    value_0 : `None | object`
        Self instance's value.
    value_1 : `None | object``
        Other instance's value.
    
    Returns
    -------
    new : `None | object`
    """
    if value_0 is None:
        return value_1
    
    if value_1 is None:
        return value_0
    
    return value_1



__all__ = ('TopGGHttpException', 'TopGGGloballyRateLimited',)

import reprlib

from .constants import JSON_KEY_EXCEPTION_MESSAGE


class TopGGGloballyRateLimited(Exception):
    """
    response : ``ClientResponse``, `None`
        Response from top.gg if received as a response.
    """
    def __init__(self, response):
        """
        Creates a new top.gg http exception instance.
        
        Parameters
        ----------
        response : ``ClientResponse``, `None`
            Response from top.gg if received as a response.
        """
        self.response = response
        Exception.__init__(self, response)


class TopGGHttpException(Exception):
    """
    Http exception raised by Top.gg.
    
    Attributes
    ----------
    response : ``ClientResponse``
        Response from top.gg.
    message : `str`
        Exception message.
    """
    def __init__(self, response, response_data):
        """
        Creates a new top.gg http exception instance.
        
        Parameters
        ----------
        response : ``ClientResponse``
            Response from top.gg.
        response_data : `str`, `dict` of (`str`, `object`) items
            Response data.
        """
        if isinstance(response_data, str):
            message = response_data
        elif isinstance(response_data, dict):
            message = response_data.get(JSON_KEY_EXCEPTION_MESSAGE, '')
        else:
            message = ''
        
        self.message = message
        self.response = response
        
        Exception.__init__(self, response, response_data)
    
    
    @property
    def status(self):
        """
        Returns the response's status code.
        
        Returns
        -------
        status_code : `int`
        """
        return self.response.status
    
    
    def __repr__(self):
        """Returns the exception's representation."""
        repr_parts = ['<', self.__class__.__name__, ' status = ', str(self.response.status)]
        
        message = self.message
        if message:
            repr_parts.append(', message = ')
            repr_parts.append(reprlib.repr(message))
        
        repr_parts.append('>')
        return ''.join(repr_parts)



__all__ = ('WelcomeScreen',)

import reprlib

from scarletio import RichAttributeErrorBaseType

from .fields import (
    parse_description, parse_welcome_channels, put_description, put_welcome_channels, validate_description,
    validate_welcome_channels
)


class WelcomeScreen(RichAttributeErrorBaseType):
    """
    Represents a guild's welcome screen.
    
    Attributes
    ----------
    description : `None`, `str`
        Description, of what is the server about.
    welcome_channels : `None`, `tuple` of ``WelcomeScreenChannel``
        The featured channels by the welcome screen.
    """
    __slots__ = ('description', 'welcome_channels', )
    
    def __new__(cls, *, description = ..., welcome_channels = ...):
        """
        Creates a new welcome screen.
        
        Parameters
        ----------
        description : `None`, `str`, Optional (Keyword only)
            Description, of what is the server about.
        welcome_channels : `None`, `iterable` of ``WelcomeScreenChannel``, Optional (Keyword only)
            The featured channels by the welcome screen.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # description
        if description is ...:
            description = None
        else:
            description = validate_description(description)
        
        # welcome_channels
        if welcome_channels is ...:
            welcome_channels = None
        else:
            welcome_channels = validate_welcome_channels(welcome_channels)
        
        # Construct
        self = object.__new__(cls)
        self.description = description
        self.welcome_channels = welcome_channels
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new welcome screen instance from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Welcome screen data.
        """
        self = object.__new__(cls)
        self.description = parse_description(data)
        self.welcome_channels = parse_welcome_channels(data)
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the welcome screen to a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields with their default values should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        data = {}
        put_description(self.description, data, defaults)
        put_welcome_channels(self.welcome_channels, data, defaults)
        return data
    
    
    def __repr__(self):
        """Returns the welcome screen's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        field_added = False
        
        # description
        description = self.description
        if (description is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' description = ')
            repr_parts.append(reprlib.repr(description))
        
        # welcome_channels
        welcome_channels = self.welcome_channels
        if (welcome_channels is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' welcome_channels = [')
            
            index = 0
            limit = len(welcome_channels)
            
            while True:
                welcome_channel = welcome_channels[index]
                
                repr_parts.append(repr(welcome_channel))
                
                index += 1
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
            
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the welcome screen's hash."""
        hash_value = 0
        
        # description
        description = self.description
        if (description is not None):
            hash_value ^= hash(description)
        
        welcome_channels = self.welcome_channels
        if (welcome_channels is not None):
            hash_value ^= len(welcome_channels)
            
            for welcome_channel in welcome_channels:
                hash_value ^= hash(welcome_channel)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two welcome screens are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.description != other.description:
            return False
        
        if self.welcome_channels != other.welcome_channels:
            return False
        
        return True
    
    
    def copy(self):
        """
        Copies the welcome screen.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.description = self.description
        welcome_channels = self.welcome_channels
        if (welcome_channels is not None):
            welcome_channels = (*welcome_channels,)
        new.welcome_channels = welcome_channels
        return new
    
    
    def copy_with(self, *, description = ..., welcome_channels = ...):
        """
        Copies the welcome screen with the given fields.
        
        Parameters
        ----------
        description : `None`, `str`, Optional (Keyword only)
            Description, of what is the server about.
        welcome_channels : `None`, `iterable` of ``WelcomeScreenChannel``, Optional (Keyword only)
            The featured channels by the welcome screen.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # description
        if description is ...:
            description = self.description
        else:
            description = validate_description(description)
        
        # welcome_channels
        if welcome_channels is ...:
            welcome_channels = self.welcome_channels
            if (welcome_channels is not None):
                welcome_channels = (*welcome_channels,)
        else:
            welcome_channels = validate_welcome_channels(welcome_channels)
        
        # Construct
        self = object.__new__(type(self))
        self.description = description
        self.welcome_channels = welcome_channels
        return self
    
    
    def iter_welcome_channels(self):
        """
        Iterates over the welcome channels of the welcome screen.
        
        This method is an iterable generator.
        
        Yields
        ------
        welcome_channel : ``WelcomeScreenChannel``
        """
        welcome_channels = self.welcome_channels
        if (welcome_channels is not None):
            yield from welcome_channels



__all__ = ('AutoModerationActionMetadataSendAlertMessage',)

from scarletio import copy_docs

from .base import AutoModerationActionMetadataBase
from .fields import parse_channel_id, put_channel_id, validate_channel_id


class AutoModerationActionMetadataSendAlertMessage(AutoModerationActionMetadataBase):
    """
    Send alert message action metadata for an auto moderation action.
    
    Attributes
    ----------
    channel_id : `int`
        The channel's identifier where the alert messages are sent.
    """
    __slots__ = ('channel_id',)
    
    def __new__(cls, channel_id = None):
        """
        Creates a new send alert message action metadata.
        
        Parameters
        ----------
        channel_id : `None`, ``Channel``, `int` = `None`, Optional
            The channel where the alert message should be sent.
               
        Raises
        ------
        TypeError
            - If `channel_id`'s type is incorrect.
        """
        channel_id = validate_channel_id(channel_id)
        
        self = object.__new__(cls)
        self.channel_id = channel_id
        return self
    
    
    @copy_docs(AutoModerationActionMetadataBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' channel_id = ')
        repr_parts.append(repr(self.channel_id))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @classmethod
    @copy_docs(AutoModerationActionMetadataBase.from_data)
    def from_data(cls, data):
        channel_id = parse_channel_id(data)
        
        self = object.__new__(cls)
        self.channel_id = channel_id
        return self
    
    
    @copy_docs(AutoModerationActionMetadataBase.to_data)
    def to_data(self, *, defaults = False):
        data = {}
        
        put_channel_id(self.channel_id, data, defaults)
        
        return data
    
    
    @copy_docs(AutoModerationActionMetadataBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self.channel_id != other.channel_id:
            return False
        
        return True
    
    
    @copy_docs(AutoModerationActionMetadataBase.__hash__)
    def __hash__(self):
        return self.channel_id
    
    
    @copy_docs(AutoModerationActionMetadataBase.copy)
    def copy(self):
        new = AutoModerationActionMetadataBase.copy(self)
        new.channel_id = self.channel_id
        return new
    
    
    def copy_with(self, *, channel_id = ...):
        """
        Copies the action metadata and modifies it's attributes by the given values.
        
        Parameters
        ----------
        channel_id : `None`, ``Channel``, `int`, Optional (Keyword only)
            The channel where the alert message should be sent.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        if channel_id is ...:
            channel_id = self.channel_id
        else:
            channel_id = validate_channel_id(channel_id)
        
        new = AutoModerationActionMetadataBase.copy(self)
        new.channel_id = channel_id
        return new



__all__ = ('ComponentType',)

from ...bases import Preinstance as P, PreinstancedBase

from ..component_metadata import (
    ComponentMetadataBase, ComponentMetadataButton, ComponentMetadataChannelSelect, ComponentMetadataMediaGallery,
    ComponentMetadataMentionableSelect, ComponentMetadataRoleSelect, ComponentMetadataRow,
    ComponentMetadataStringSelect, ComponentMetadataSeparator, ComponentMetadataText, ComponentMetadataTextInput,
    ComponentMetadataUserSelect
)

from .flags import ComponentTypeLayoutFlag


COMPONENT_TYPE_LAYOUT_FLAGS_ALL = ComponentTypeLayoutFlag().update_by_keys(top_level = True, nestable = True)


class ComponentType(PreinstancedBase, value_type = int):
    """
    Represents a component's type.
    
    Attributes
    ----------
    layout_flags : ``ComponentTypeLayoutFlag``
        Flags about the component's layout information.
    
    metadata_type : `type<ComponentMetadataBase>`
        Metadata type.
    
    name : `str`
        The name of the component type.
    
    value : `int`
        The identifier value the component type.
    
    Type Attributes
    ---------------
    Every predefined component type can be accessed as type attribute as well:
    
    +-----------------------+-------------------------------+-------+
    | Type attribute name   | Name                          | Value |
    +=======================+===============================+=======+
    | none                  | none                          | 0     |
    +-----------------------+-------------------------------+-------+
    | row                   | row                           | 1     |
    +-----------------------+-------------------------------+-------+
    | button                | button                        | 2     |
    +-----------------------+-------------------------------+-------+
    | string_select         | string select                 | 3     |
    +-----------------------+-------------------------------+-------+
    | text_input            | text input                    | 4     |
    +-----------------------+-------------------------------+-------+
    | user_select           | user select                   | 5     |
    +-----------------------+-------------------------------+-------+
    | role_select           | role select                   | 6     |
    +-----------------------+-------------------------------+-------+
    | mentionable_select    | mentionable select            | 7     |
    +-----------------------+-------------------------------+-------+
    | channel_select        | channel select                | 8     |
    +-----------------------+-------------------------------+-------+
    | ???                   | ???                           | 9     |
    +-----------------------+-------------------------------+-------+
    | text                  | text                          | 10    |
    +-----------------------+-------------------------------+-------+
    | ???                   | ???                           | 11    |
    +-----------------------+-------------------------------+-------+
    | media_gallery         | media gallery                 | 12    |
    +-----------------------+-------------------------------+-------+
    | ???                   | ???                           | 13    |
    +-----------------------+-------------------------------+-------+
    | separator             | separator                     | 14    |
    +-----------------------+-------------------------------+-------+
    | ???                   | ???                           | 15    |
    +-----------------------+-------------------------------+-------+
    | ???                   | activity content inventory    | 16    |
    +-----------------------+-------------------------------+-------+
    """
    __slots__ = ('layout_flags', 'metadata_type', )
    
    
    def __new__(cls, value, name= None, metadata_type = None, layout_flags = ...):
        """
        Creates a new component type.
        
        Parameters
        ----------
        value : `int`
            The Discord side identifier value of the channel type.
        
        name : `None | str` = `None`, Optional
            The default name of the channel type.
        
        metadata_type : `None | type<ComponentMetadataBase>` = `None`, Optional
            The component type's respective metadata type.
        
        layout_flags : ``ComponentTypeLayoutFlag``, Optional
            Flags about the component's layout information.
        """
        if metadata_type is None:
            metadata_type = ComponentMetadataBase
        
        if layout_flags is ...:
            layout_flags = COMPONENT_TYPE_LAYOUT_FLAGS_ALL
        
        self = PreinstancedBase.__new__(cls, value, name)
        self.layout_flags = layout_flags
        self.metadata_type = metadata_type
        return self
    
    
    none = P(
        0,
        'none',
        ComponentMetadataBase,
        ComponentTypeLayoutFlag(),
    )
    row = P(
        1,
        'row',
        ComponentMetadataRow,
        ComponentTypeLayoutFlag().update_by_keys(top_level = True),
    )
    button = P(
        2,
        'button',
        ComponentMetadataButton,
        ComponentTypeLayoutFlag().update_by_keys(nestable = True),
    )
    string_select = P(
        3,
        'string select',
        ComponentMetadataStringSelect,
        ComponentTypeLayoutFlag().update_by_keys(nestable = True),
    )
    text_input = P(
        4,
        'text input',
        ComponentMetadataTextInput,
        ComponentTypeLayoutFlag().update_by_keys(nestable = True),
    )
    user_select = P(
        5,
        'user select',
        ComponentMetadataUserSelect,
        ComponentTypeLayoutFlag().update_by_keys(nestable = True),
    )
    role_select = P(
        6,
        'role select',
        ComponentMetadataRoleSelect,
        ComponentTypeLayoutFlag().update_by_keys(nestable = True),
    )
    mentionable_select = P(
        7,
        'mentionable select',
        ComponentMetadataMentionableSelect,
        ComponentTypeLayoutFlag().update_by_keys(nestable = True),
    )
    channel_select = P(
        8,
        'channel select',
        ComponentMetadataChannelSelect,
        ComponentTypeLayoutFlag().update_by_keys(nestable = True),
    )
    
    # Not released, so we do not know their layout
    text = P(
        10,
        'text',
        ComponentMetadataText,
        COMPONENT_TYPE_LAYOUT_FLAGS_ALL,
    )
    media_gallery = P(
        12,
        'media gallery',
        ComponentMetadataMediaGallery,
        COMPONENT_TYPE_LAYOUT_FLAGS_ALL,
    )
    separator = P(
        14,
        'separator',
        ComponentMetadataSeparator,
        COMPONENT_TYPE_LAYOUT_FLAGS_ALL,
    )



from .base import *
from .channel import *
from .constants import *
from .fields import *
from .float import *
from .integer import *
from .nested import *
from .numeric import *
from .parameter import *
from .primitive import *
from .string import *
from .sub_command import *


__all__ = (
    *base.__all__,
    *channel.__all__,
    *constants.__all__,
    *fields.__all__,
    *float.__all__,
    *integer.__all__,
    *nested.__all__,
    *numeric.__all__,
    *parameter.__all__,
    *primitive.__all__,
    *string.__all__,
    *sub_command.__all__,
)



__all__ = ('InteractionResponseContext',)

from scarletio import Task, to_coroutine

from ...core import KOKORO

from .constants import (
    RESPONSE_FLAG_ACKNOWLEDGED, RESPONSE_FLAG_ACKNOWLEDGING_OR_ACKNOWLEDGED, RESPONSE_FLAG_DEFERRED,
    RESPONSE_FLAG_DEFERRING, RESPONSE_FLAG_EPHEMERAL, RESPONSE_FLAG_RESPONDED, RESPONSE_FLAG_RESPONDING,
    RESPONSE_FLAG_RESPONDING_OR_RESPONDED
)


class InteractionResponseContext:
    """
    Interaction response context manager for managing the interaction event's response flag.
    
    Attributes
    ----------
    interaction_event : ``InteractionEvent``
        The respective interaction event event.
    
    deferring : `bool`
        Whether the request just deferring the interaction event.
    
    ephemeral : `bool`
        Whether the request is ephemeral.
    """
    __slots__ = ('interaction_event', 'deferring', 'ephemeral',)
    
    def __new__(cls, interaction_event, deferring, ephemeral):
        """
        Creates a new ``InteractionResponseContext`` with the given parameters.
        
        Parameters
        ----------
        interaction_event : ``InteractionEvent``
            The interaction event to which wew are responding to.
        
        deferring : `bool`
            Whether the request just deferring the interaction event.
        
        ephemeral : `bool`
            Whether the request is ephemeral.
        """
        self = object.__new__(cls)
        self.interaction_event = interaction_event
        self.deferring = deferring
        self.ephemeral = ephemeral
        return self
    
    
    def __repr__(self):
        """Returns the interaction response context's representation."""
        repr_parts = ['<', type(self).__name__]
        
        if self.deferring:
            repr_parts.append(' deferring')
        
        if self.ephemeral:
            repr_parts.append(' (ephemeral)')
        
        repr_parts.append(' interaction_event = ')
        repr_parts.append(repr(self.interaction_event))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the interaction response context's hash value."""
        hash_value = 0
        
        # deferring
        hash_value ^= self.deferring << 2
        
        # ephemeral
        hash_value ^= self.ephemeral << 6
        
        # interaction_event
        hash_value ^= hash(self.interaction_event)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two interaction response contexts are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # deferring
        if self.deferring != other.deferring:
            return False
        
        # ephemeral
        if self.ephemeral != other.ephemeral:
            return False
        
        # interaction_event
        if self.interaction_event != other.interaction_event:
            return False
        
        return True
    
    
    @to_coroutine
    def ensure(self, coroutine, callback = None):
        """
        Ensures the coroutine within the interaction response context
        
        This method is an awaitable generator.
        
        Parameters
        ----------
        coroutine : `GeneratorType | CoroutineType`
            Coroutine to run.
        
        callback : `None | FunctionType` = `None`, Optional
            Additional function to execute after the coroutine.
        """
        self.interaction_event._async_task = Task(KOKORO, self._async(coroutine, callback))
        yield # skip a ready cycle
    
    
    async def _async(self, coroutine, callback):
        """
        Ensures the coroutine within the interaction response context. Wrapped into a task by ``.ensure``.
        
        This method is a coroutine.
        
        Parameters
        ----------
        coroutine : `GeneratorType | CoroutineType`
            Coroutine to run.
        
        callback : `None | FunctionType`
            Additional function to execute after the coroutine.
        """
        try:
            async with self:
                interaction_response_data = await coroutine
        finally:
            self.interaction_event._async_task = None
        
        if (callback is not None):
            callback(self.interaction_event, interaction_response_data)
    
    
    async def __aenter__(self):
        """
        Enters the context manager as deferring or responding if applicable.
        
        This method is a coroutine.
        """
        interaction_event = self.interaction_event
        await interaction_event._wait_for_async_task_completion()
        
        response_flag = interaction_event._response_flags
        
        if self.deferring:
            if not (response_flag & RESPONSE_FLAG_ACKNOWLEDGING_OR_ACKNOWLEDGED):
                response_flag |= RESPONSE_FLAG_DEFERRING
        else:
            if (not response_flag & RESPONSE_FLAG_RESPONDING_OR_RESPONDED):
                response_flag |= RESPONSE_FLAG_RESPONDING
        
        interaction_event._response_flags = response_flag
        
        return self
    
    
    async def __aexit__(self, exception_type, exception_value, exception_traceback):
        """
        Exits the context manager, marking the interaction event as deferred or responded if no exception occurred.
        
        This method is a coroutine.
        """
        interaction_event = self.interaction_event
        response_flag = interaction_event._response_flags
        if exception_type is None:
            if self.ephemeral:
                if not response_flag & RESPONSE_FLAG_ACKNOWLEDGED:
                    response_flag |= RESPONSE_FLAG_EPHEMERAL
            
            if self.deferring:
                if response_flag & RESPONSE_FLAG_DEFERRING:
                    response_flag ^= RESPONSE_FLAG_DEFERRING
                    response_flag |= RESPONSE_FLAG_DEFERRED
            else:
                if response_flag & RESPONSE_FLAG_RESPONDING:
                    response_flag ^= RESPONSE_FLAG_RESPONDING
                    response_flag |= RESPONSE_FLAG_RESPONDED
                    
                    if response_flag & RESPONSE_FLAG_DEFERRED:
                        response_flag ^= RESPONSE_FLAG_DEFERRED
        
        else:
            if self.deferring:
                if response_flag & RESPONSE_FLAG_DEFERRING:
                    response_flag ^= RESPONSE_FLAG_DEFERRING
            else:
                if response_flag & RESPONSE_FLAG_RESPONDING:
                    response_flag ^= RESPONSE_FLAG_RESPONDING
        
        interaction_event._response_flags = response_flag
        return False



__all__ = ('MessageCall', )

from scarletio import RichAttributeErrorBaseType

from ...user import create_partial_user_from_id
from ...utils import DATETIME_FORMAT_CODE
from .fields import (
    parse_ended_at, parse_user_ids, put_ended_at, put_user_ids, validate_ended_at, validate_user_ids
)


class MessageCall(RichAttributeErrorBaseType):
    """
    Contained by messages with message type `call`.
    
    Attributes
    ----------
    ended_at : `None`, `datetime`
        When the call ended. Set as `None` if it is still ongoing.
    user_ids : `None`, `tuple` of `int`
        The users' identifiers participating the call.
    """
    __slots__ = ('ended_at', 'user_ids',)
    
    def __new__(cls, *, ended_at = ..., user_ids = ...):
        """
        Creates a new message call from the given parameters.
        
        Parameters
        ----------
        ended_at : `None`, `datetime`, Optional (Keyword only)
            When the call ended.
        user_ids : `None`, `iterable` of (`ClientUserBase`, `int`), Optional (Keyword only)
            The users' identifiers participating the call.
        
        Raises
        ------
        TypeError
            - If a parameter's user_ids is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # ended_at
        if ended_at is ...:
            ended_at = None
        else:
            ended_at = validate_ended_at(ended_at)
        
        # user_ids
        if user_ids is ...:
            user_ids = None
        else:
            user_ids = validate_user_ids(user_ids)
        
        # Construct
        self = object.__new__(cls)
        self.ended_at = ended_at
        self.user_ids = user_ids
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new message call instance instance from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Message call data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.ended_at = parse_ended_at(data)
        self.user_ids = parse_user_ids(data)
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the message call back to json a serializable dictionary.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`)
        """
        data = {}
        put_ended_at(self.ended_at, data, defaults)
        put_user_ids(self.user_ids, data, defaults)
        return data
    
    
    def __eq__(self, other):
        """Returns whether the two message calls are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # ended_at
        if self.ended_at != other.ended_at:
            return False
        
        # user_ids
        if self.user_ids != other.user_ids:
            return False
        
        return True
    
    
    def __repr__(self):
        """Returns the message call's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        field_added = False
        
        user_ids = self.user_ids
        repr_parts.append(' user_ids = [')
        
        if (user_ids is not None):
            index = 0
            length = len(user_ids)
            
            while True:
                user_id = user_ids[index]
                repr_parts.append(repr(user_id))
                index += 1
                if index == length:
                    break
                
                repr_parts.append(', ')
                continue
        
        repr_parts.append(']')
        
        ended_at = self.ended_at
        if (ended_at is not None):
            repr_parts.append(', ended_at = ')
            repr_parts.append(format(ended_at, DATETIME_FORMAT_CODE))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the message call's hash value."""
        hash_value = 0
        
        # ended_at
        ended_at = self.ended_at
        if (ended_at is not None):
            hash_value ^= hash(ended_at)
        
        # user_ids
        user_ids = self.user_ids
        if (user_ids is not None):
            hash_value ^= len(user_ids)
            
            for user_id in user_ids:
                hash_value ^= user_id
        
        return hash_value
    
    
    def copy(self):
        """
        Copies the message call.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.ended_at = self.ended_at
        
        user_ids = self.user_ids
        if (user_ids is not None):
            user_ids = (*user_ids,)
        new.user_ids = user_ids
        
        return new
    
    
    def copy_with(self, *, user_ids = ..., ended_at = ...):
        """
        Copies the message call with the given fields.
        
        Parameters
        ----------
        ----------
        ended_at : `None`, `datetime`, Optional (Keyword only)
            When the call ended.
        user_ids : `None`, `iterable` of (`ClientUserBase`, `int`), Optional (Keyword only)
            The users' identifiers participating the call.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's user_ids is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # ended_at
        if ended_at is ...:
            ended_at = self.ended_at
        else:
            ended_at = validate_ended_at(ended_at)
        
        # user_ids
        if user_ids is ...:
            user_ids = self.user_ids
            if (user_ids is not None):
                user_ids = (*user_ids,)
        else:
            user_ids = validate_user_ids(user_ids)
        
        # Construct
        new = object.__new__(type(self))
        new.ended_at = ended_at
        new.user_ids = user_ids
        return new
    
    
    def iter_user_ids(self):
        """
        Iterates over the users' identifiers who participate (or participated) in the call.
        
        This method is an iterable generator.
        
        Yields
        ------
        user_id : `int`
        """
        user_ids = self.user_ids
        if (user_ids is not None):
            yield from user_ids
    
    
    def users(self):
        """
        Returns the users who participate (or participated) in the call.
        
        Returns
        -------
        users : `None`, `tuple` of ``ClientUserBase``
        """
        user_ids = self.user_ids
        if (user_ids is not None):
            return (*(create_partial_user_from_id(user_id) for user_id in user_ids),)
    
    
    def iter_users(self):
        """
        Iterates over the users who participate (or participated) in the call.
        
        This method is an iterable generator.
        
        Yields
        ------
        user : ``ClientUserBase``
        """
        user_ids = self.user_ids
        if (user_ids is not None):
            for user_id in user_ids:
                yield create_partial_user_from_id(user_id)



__all__ = ()

from ...channel import (
    ChannelFlag, ChannelType, ForumLayout, ForumTag, PermissionOverwrite, SortOrder, VideoQualityMode, VoiceRegion
)
from ...channel.channel.fields import validate_type
from ...channel.channel_metadata.constants import (
    AUTO_ARCHIVE_DEFAULT, BITRATE_DEFAULT, SLOWMODE_DEFAULT, USER_LIMIT_DEFAULT
)
from ...channel.channel_metadata.fields import (
    validate_applied_tag_ids, validate_archived, validate_auto_archive_after, validate_available_tags, validate_bitrate,
    validate_default_forum_layout, validate_default_sort_order, validate_default_thread_auto_archive_after,
    validate_default_thread_reaction_emoji, validate_default_thread_slowmode, validate_flags, validate_invitable,
    validate_name, validate_nsfw, validate_open, validate_parent_id, validate_permission_overwrites, validate_position,
    validate_region, validate_slowmode, validate_topic, validate_user_limit, validate_video_quality_mode
)
from ...emoji import create_partial_emoji_data, create_partial_emoji_from_data

from ..audit_log_entry_change_conversion import AuditLogEntryChangeConversion, AuditLogEntryChangeConversionGroup
from ..audit_log_entry_change_conversion.change_deserializers import change_deserializer_deprecation
from ..conversion_helpers.converters import (
    value_deserializer_description, value_deserializer_id, value_deserializer_ids, value_deserializer_name,
    value_serializer_description, value_serializer_id, value_serializer_ids, value_serializer_name
)


# ---- applied_tag_ids ----

APPLIED_TAG_IDS_CONVERSION = AuditLogEntryChangeConversion(
    ('applied_tags',),
    'applied_tag_ids',
    value_deserializer = value_deserializer_ids,
    value_serializer = value_serializer_ids,
    value_validator = validate_applied_tag_ids,
)


# ---- archived ----

ARCHIVED_CONVERSION = AuditLogEntryChangeConversion(
    ('archived',),
    'archived',
    value_validator = validate_archived,
)


@ARCHIVED_CONVERSION.set_value_deserializer
def archived_value_deserializer(value):
    if value is None:
        value = False

    return value


# ---- auto_archive_after ----

AUTO_ARCHIVE_AFTER_CONVERSION = AuditLogEntryChangeConversion(
    ('auto_archive_duration',),
    'auto_archive_after',
    value_validator = validate_auto_archive_after,
)


@AUTO_ARCHIVE_AFTER_CONVERSION.set_value_deserializer
def auto_archive_after_value_deserializer(value):
    if value is None:
        value = AUTO_ARCHIVE_DEFAULT
    else:
        value *= 60
    return value


@AUTO_ARCHIVE_AFTER_CONVERSION.set_value_serializer
def auto_archive_after_value_serializer(value):
    return value // 60


# --- available_tags ----

AVAILABLE_TAGS_CONVERSION = AuditLogEntryChangeConversion(
    ('available_tags',),
    'available_tags',
    value_validator = validate_available_tags,
)


@AVAILABLE_TAGS_CONVERSION.set_value_deserializer
def available_tags_value_deserializer(value):
    if value is None:
        pass
    elif (not value):
        value = None
    else:
        value = (*sorted(ForumTag.from_data(tag_data) for tag_data in value),)
    return value


@AVAILABLE_TAGS_CONVERSION.set_value_serializer
def available_tags_value_serializer(value):
    if value is None:
        value = []
    else:
        value = [forum_tag.to_data(defaults = True, include_internals = True) for forum_tag in value]
    return value


# ---- bitrate ----

BITRATE_CONVERSION = AuditLogEntryChangeConversion(
    ('bitrate',),
    'bitrate',
    value_validator = validate_bitrate,
)


@BITRATE_CONVERSION.set_value_deserializer
def bitrate_value_deserializer(value):
    if value is None:
        value = BITRATE_DEFAULT
    return value


# ---- default_thread_auto_archive_after ----

DEFAULT_THREAD_AUTO_ARCHIVE_AFTER_CONVERSION = AuditLogEntryChangeConversion(
    ('default_auto_archive_duration',),
    'default_thread_auto_archive_after',
    value_validator = validate_default_thread_auto_archive_after,
)


@DEFAULT_THREAD_AUTO_ARCHIVE_AFTER_CONVERSION.set_value_deserializer
def default_thread_auto_archive_after_value_deserializer(value):
    if value is None:
        value = AUTO_ARCHIVE_DEFAULT
    else:
        value *= 60
    return value


@DEFAULT_THREAD_AUTO_ARCHIVE_AFTER_CONVERSION.set_value_serializer
def default_thread_auto_archive_after_value_serializer(value):
    return value // 60


# ---- default_forum_layout ----

DEFAULT_FORUM_LAYOUT_CONVERSION = AuditLogEntryChangeConversion(
    ('default_forum_layout',),
    'default_forum_layout',
    value_validator = validate_default_forum_layout,
)


@DEFAULT_FORUM_LAYOUT_CONVERSION.set_value_deserializer
def default_forum_layout_value_deserializer(value):
    return ForumLayout(value)


@DEFAULT_FORUM_LAYOUT_CONVERSION.set_value_serializer
def default_forum_layout_value_serializer(value):
    return value.value


# ---- default_sort_order ----

DEFAULT_SORT_ORDER_CONVERSION = AuditLogEntryChangeConversion(
    ('default_sort_order',),
    'default_sort_order',
    value_validator = validate_default_sort_order,
)


@DEFAULT_SORT_ORDER_CONVERSION.set_value_deserializer
def default_sort_order_value_deserializer(value):
    return SortOrder(value)


@DEFAULT_SORT_ORDER_CONVERSION.set_value_serializer
def default_sort_order_value_serializer(value):
    return value.value


# ---- default_thread_reaction_emoji ----

DEFAULT_THREAD_REACTION_EMOJI_CONVERSION = AuditLogEntryChangeConversion(
    ('default_reaction_emoji',),
    'default_thread_reaction_emoji',
    value_validator = validate_default_thread_reaction_emoji,
)


@DEFAULT_THREAD_REACTION_EMOJI_CONVERSION.set_value_deserializer
def default_thread_reaction_emoji_value_deserializer(value):
    if (value is not None):
        value = create_partial_emoji_from_data(value)
    
    return value


@DEFAULT_THREAD_REACTION_EMOJI_CONVERSION.set_value_serializer
def default_thread_reaction_emoji_value_serializer(value):
    if value is not None:
        value = create_partial_emoji_data(value)
    
    return value


# ---- default_thread_slowmode ----

DEFAULT_THREAD_SLOWMODE_CONVERSION = AuditLogEntryChangeConversion(
    ('default_thread_rate_limit_per_user',),
    'default_thread_slowmode',
    value_validator = validate_default_thread_slowmode,
)

@DEFAULT_THREAD_SLOWMODE_CONVERSION.set_value_deserializer
def default_thread_slowmode_value_deserializer(value):
    if value is None:
        value = SLOWMODE_DEFAULT
    return value


# ---- flags ----

FLAGS_CONVERSION = AuditLogEntryChangeConversion(
    ('flags',),
    'flags',
    value_validator = validate_flags,
)


@FLAGS_CONVERSION.set_value_deserializer
def flags_value_deserializer(value):
    if value is None:
        value = ChannelFlag()
    else:
        value = ChannelFlag(value)
    
    return value


@FLAGS_CONVERSION.set_value_serializer
def flags_value_serializer(value):
    return int(value)


# ---- icon_emoji  ----

ICON_EMOJI_CONVERSION_IGNORED = AuditLogEntryChangeConversion(
    ('icon_emoji',),
    '',
    change_deserializer = change_deserializer_deprecation,
)


# ---- invitable ----

INVITABLE_CONVERSION = AuditLogEntryChangeConversion(
    ('invitable',),
    'invitable',
    value_validator = validate_invitable,
)


@INVITABLE_CONVERSION.set_value_deserializer
def invitable_value_deserializer(value):
    if value is None:
        value = True
    return value


# ---- name ----

NAME_CONVERSION = AuditLogEntryChangeConversion(
    ('name',),
    'name',
    value_deserializer = value_deserializer_name,
    value_serializer = value_serializer_name,
    value_validator = validate_name,
)


# ---- nsfw ----

NSFW_CONVERSION = AuditLogEntryChangeConversion(
    ('nsfw',),
    'nsfw',
    value_validator = validate_nsfw,
)


@NSFW_CONVERSION.set_value_deserializer
def nsfw_value_deserializer(value):
    if value is None:
        value = False
    return value


# ---- open ----

OPEN_CONVERSION = AuditLogEntryChangeConversion(
    ('locked',),
    'open',
    value_validator = validate_open,
)


@OPEN_CONVERSION.set_value_deserializer
def open_value_deserializer(value):
    if value is None:
        value = True
    else:
        value = not value
    return value


@OPEN_CONVERSION.set_value_serializer
def open_value_serializer(value):
    return not value


# ---- parent_id ----

PARENT_ID_CONVERSION = AuditLogEntryChangeConversion(
    ('parent_id',),
    'parent_id',
    value_deserializer = value_deserializer_id,
    value_serializer = value_serializer_id,
    value_validator = validate_parent_id,
)


# ---- permission_overwrites ----

PERMISSION_OVERWRITES_CONVERSION = AuditLogEntryChangeConversion(
    ('permission_overwrites',),
    'permission_overwrites',
    value_validator = validate_permission_overwrites,
)

@PERMISSION_OVERWRITES_CONVERSION.set_value_deserializer
def permission_overwrite_value_deserializer(value):
    if value is None:
        pass
    elif (not value):
        value = None
    else:
        value = {
            permission_overwrite.target_id: permission_overwrite for permission_overwrite in
            (PermissionOverwrite.from_data(overwrite_data) for overwrite_data in value)
        }
    
    return value


@PERMISSION_OVERWRITES_CONVERSION.set_value_serializer
def permission_overwrite_value_serializer(value):
    if value is None:
        value = []
    else:
        value = [
            permission_overwrite.to_data(defaults = True, include_internals = True)
            for permission_overwrite in sorted(value.values())
        ]
    
    return value


# ---- position ----

POSITION_CONVERSION = AuditLogEntryChangeConversion(
    ('position',),
    'position',
    value_validator = validate_position,
)

@POSITION_CONVERSION.set_value_deserializer
def position_value_deserializer(value):
    if value is None:
        value = 0
    return value


# ---- region ----

REGION_CONVERSION = AuditLogEntryChangeConversion(
    ('rtc_region',),
    'region',
    value_validator = validate_region,
)


@REGION_CONVERSION.set_value_deserializer
def region_value_deserializer(value):
    return VoiceRegion(value)


@REGION_CONVERSION.set_value_serializer
def region_value_serializer(value):
    return value.value


# ---- slowmode ----

SLOWMODE_CONVERSION = AuditLogEntryChangeConversion(
    ('rate_limit_per_user',),
    'slowmode',
    value_validator = validate_slowmode,
)

@SLOWMODE_CONVERSION.set_value_deserializer
def slowmode_value_deserializer(value):
    if value is None:
        value = SLOWMODE_DEFAULT
    return value


# ---- template  ----

TEMPLATE_CONVERSION_IGNORED = AuditLogEntryChangeConversion(
    ('template',),
    '',
    change_deserializer = change_deserializer_deprecation,
)


# ---- theme_color  ----

THEME_COLOR_CONVERSION_IGNORED = AuditLogEntryChangeConversion(
    ('theme_color',),
    '',
    change_deserializer = change_deserializer_deprecation,
)


# ---- topic ----

TOPIC_CONVERSION = AuditLogEntryChangeConversion(
    ('topic',),
    'topic',
    value_deserializer = value_deserializer_description,
    value_serializer = value_serializer_description,
    value_validator = validate_topic,
)


# ---- type ----

TYPE_CONVERSION = AuditLogEntryChangeConversion(
    ('type',),
    'type',
    value_validator = validate_type,
)


@TYPE_CONVERSION.set_value_deserializer
def type_value_deserializer(value):
    return ChannelType(value)


@TYPE_CONVERSION.set_value_serializer
def type_value_serializer(value):
    return value.value


# ---- video_quality_mode ----

VIDEO_QUALITY_MODE_CONVERSION = AuditLogEntryChangeConversion(
    ('video_quality_mode',),
    'video_quality_mode',
    value_validator = validate_video_quality_mode,
)


@VIDEO_QUALITY_MODE_CONVERSION.set_value_deserializer
def video_quality_mode_value_deserializer(value):
    return VideoQualityMode(value)


@VIDEO_QUALITY_MODE_CONVERSION.set_value_serializer
def video_quality_mode_value_serializer(value):
    return value.value


# ---- user_limit ----

USER_LIMIT_CONVERSION = AuditLogEntryChangeConversion(
    ('user_limit',),
    'user_limit',
    value_validator = validate_user_limit,
)

@USER_LIMIT_CONVERSION.set_value_deserializer
def user_limit_value_deserializer(value):
    if value is None:
        value = USER_LIMIT_DEFAULT
    return value


# ---- Construct ----

CHANNEL_CONVERSIONS = AuditLogEntryChangeConversionGroup(
    APPLIED_TAG_IDS_CONVERSION,
    ARCHIVED_CONVERSION,
    AUTO_ARCHIVE_AFTER_CONVERSION,
    AVAILABLE_TAGS_CONVERSION,
    BITRATE_CONVERSION,
    DEFAULT_THREAD_AUTO_ARCHIVE_AFTER_CONVERSION,
    DEFAULT_FORUM_LAYOUT_CONVERSION,
    DEFAULT_SORT_ORDER_CONVERSION,
    DEFAULT_THREAD_REACTION_EMOJI_CONVERSION,
    DEFAULT_THREAD_SLOWMODE_CONVERSION,
    FLAGS_CONVERSION,
    ICON_EMOJI_CONVERSION_IGNORED,
    INVITABLE_CONVERSION,
    NAME_CONVERSION,
    NSFW_CONVERSION,
    OPEN_CONVERSION,
    PARENT_ID_CONVERSION,
    PERMISSION_OVERWRITES_CONVERSION,
    POSITION_CONVERSION,
    REGION_CONVERSION,
    SLOWMODE_CONVERSION,
    TEMPLATE_CONVERSION_IGNORED,
    THEME_COLOR_CONVERSION_IGNORED,
    TOPIC_CONVERSION,
    TYPE_CONVERSION,
    VIDEO_QUALITY_MODE_CONVERSION,
    USER_LIMIT_CONVERSION,
)



__all__ = ()

from scarletio import RichAttributeErrorBaseType

from .constants import LATENCY_DEFAULT


class DiscordGatewayBase(RichAttributeErrorBaseType):
    """
    Base type for gateways used by ``Client``-s to communicate with Discord with web socket.
    """
    __slots__ = ()
    
    def __new__(cls):
        """
        Creates a new gateway
        """
        return object.__new__(cls)
    
    
    def __repr__(self):
        """Returns the gateway's representation."""
        repr_parts = ['<', type(self).__name__]
        self._put_repr_parts_into(repr_parts)
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def _put_repr_parts_into(self, repr_parts):
        """
        Helper method to extend `repr_parts` with type specific fields.
        """
        pass
    
    
    async def run(self):
        """
        Keeps the gateway receiving message and processing it. If the gateway needs to be reconnected, reconnects
        itself. If connecting cannot succeed, because there is no internet returns `True`. If the `.client` is
        stopped, then returns `False`.
        
        If `True` is returned the respective client stops all other gateways as well and tries to reconnect. When
        the internet is back the client will launch back the gateway.
        
        This method is a coroutine.
        
        Returns
        -------
        outcome : `bool<False>`
            Always `False`.
        
        Raises
        ------
        DiscordGatewayException
            The client tries to connect with bad or not acceptable intent or shard value.
        DiscordException
        """
        raise NotImplementedError
    
    
    @property
    def latency(self):
        """
        The latency of the web socket in seconds.
        If no latency is recorded will return the default latency.
        
        Returns
        -------
        latency : `float`
        """
        return LATENCY_DEFAULT

    
    async def terminate(self):
        """
        Terminates the gateway's beating and closes it's web socket with close code of `4000`.
        
        This method is a coroutine.
        """
        pass
    
    
    async def close(self):
        """
        Cancels the gateway's heartbeat and closes it's web socket with close code of `1000`.
        
        This method is a coroutine.
        """
        pass
    
    
    def abort(self):
        """
        Abort the gateway immediately.
        """
        pass
    
    
    async def send_as_json(self, data):
        """
        Sends the data as json to Discord on the gateway's web socket.
        If there is no web socket, or the web socket is closed will not raise.
        
        This method is a coroutine.
        
        Parameters
        ----------
        data : `object`
            The data to send.
        """
        pass
    
    
    async def beat(self):
        """
        Sends a heartbeat packet to Discord.
        
        This method is a coroutine.
        """
        pass



__all__ = ('CONVERSION_ATTACHMENTS',)

from collections import deque as Deque
from os.path import split as split_path

from scarletio.web_common import FormData

from .....env import API_VERSION

from ....builder.constants import CONVERSION_KIND_FIELD
from ....builder.conversion import Conversion
from ....utils import random_id

from ...attachment import Attachment


def _is_attachments(value):
    """
    Yields the outcome if the `value` is acceptable.
    
    This function is a generator.
    
    Parameters
    ----------
    value : `None`, `(str, object)`, `dict<str, object>, \
            `(list | Deque)<(object,) | (None | str, object) | (None | str, object, None | str)>)`
        The value to check
    
    Yields
    ------
    attachments : `None | list<(bool<True>, int) | (bool<False>, (str, object, None | str))>`
        The processed attachments
    """
    # None
    if value is None:
        yield None
        return
    
    # tuple
    if isinstance(value, tuple):
        for attachment in _is_valid_tuple_attachment(value):
            yield [attachment]
        return
    
    # Attachment
    if isinstance(value, Attachment):
        yield [(True, value.id)]
        return
    
    # list | Deque
    if isinstance(value, list) or isinstance(value, Deque):
        attachments = None
        
        for element in value:
            for attachment in _is_single_attachment(element):
                break
            else:
                return
            
            if attachments is None:
                attachments = []
            
            attachments.append(attachment)
            continue
        
        yield attachments
        return
    
    # dict-like
    if hasattr(type(value), 'items'):
        attachments = None
        
        for item in value.items():
            if attachments is None:
                attachments = []
            
            attachments.append((False, (*item, None)))
        
        yield attachments
        return
    
    # rest
    yield [(False, (_get_or_create_io_name(value), value, None))]
    return


def _is_single_attachment(value):
    """
    Yields the outcome if the `value` is acceptable.
    
    This function is a generator.
    
    Parameters
    ----------
    value : `None`, `(str, object)`, `dict<str, object>, \
            `(list | Deque)<(object,) | (None | str, object) | (None | str, object, None | str)>)`
        The value to check
    
    Yields
    ------
    attachments : `(bool<True>, int) | (bool<False>, (str, object, None | str))`
        The processed attachments
    """
    if isinstance(value, tuple):
        yield from _is_valid_tuple_attachment(value)
        return
    
    if isinstance(value, Attachment):
        yield (True, value.id)
        return
    
    yield (False, (_get_or_create_io_name(value), value, None))
    return


def _get_or_create_io_name(io):
    """
    Gets the `io`'s name or creates a new one.
    
    Parameters
    ----------
    io : `object`
        Io to get its name of.
    
    Returns
    -------
    mame : `str`
    """
    name = getattr(io, 'name', None)
    if (name is not None) and name:
        name = split_path(name)[1]
    else:
        name = str(random_id())
    
    return name


def _is_valid_tuple_attachment(value):
    """
    Returns whether the given `tuple` is a valid attachment.
    
    This function is a generator.
    
    Parameters
    ----------
    value : `(None | str, object) | (None | str, object, None | str)`
        The value to check.
    
    Yields
    ------
    attachment : `(bool<False>, (str, object, None | str))`
    """
    length = len(value)
    if length < 1 or length > 3:
        return
        
    if length == 1:
        io, = value
        name = None
        description = None
    elif length == 2:
        name, io = value
        if (name is not None) and (not name):
            name = None
        description = None
    else:
        name, io, description = value
        if (name is not None) and (not name):
            name = None
        if (description is not None) and (not description):
            description = None
    
    if (name is None) or (not name):
        name = _get_or_create_io_name(io)
    
    yield False, (name, io, description)


def _build_partial_attachment_data(attachment_id, description):
    """
    Builds a partial attachment data to be sent to Discord.
    
    Parameters
    ----------
    attachment_id : `int`
        The attachment's identifier.
    description : `None | str`
        Description for the attachment.
    
    Returns
    -------
    data : `dict<str, object>
    """
    data = {'id': str(attachment_id)}
    
    if (description is not None):
        data['description'] = description
    
    return data


class CONVERSION_ATTACHMENTS(Conversion):
    # Generic
    
    name = 'attachments'
    name_aliases = ['files', 'file']
    expected_types_messages = (
        '`None`, `(str, object)`, `dict<str, object>, '
        '`(list | Deque)<(object,) | (None | str, object) | (None | str, object, None | str)>)`'
    )
    kind = CONVERSION_KIND_FIELD
    output_conversion = None
    
    # Setting
    
    set_merger = None
    set_type = None
    set_type_processor = None
    set_identifier = None
    set_listing_identifier = None
    
    set_validator = _is_attachments
    
    
    # Reading
    
    get_default = None
    get_processor = None
    
    
    # Serialization
    
    serializer_key = None
    serializer_optional = None
    serializer_required = None
    
    
    if API_VERSION >= 9:
        def serializer_putter(data, defaults, value):
            
            file_attachment_index = 0
            
            if value is None:
                attachment_datas = None
            
            else:
                attachment_datas = []
                
                for is_attachment, details in value:
                    if is_attachment:
                        attachment_id = details
                        description = None
                    else:
                        attachment_id = file_attachment_index
                        file_attachment_index += 1
                        description = details[2]
                    
                    attachment_datas.append(_build_partial_attachment_data(attachment_id, description))
            
            
            if (not defaults) and (attachment_datas is None):
                return data
            
            if attachment_datas is None:
                data['attachments'] = []
            else:
                data['attachments'] = attachment_datas
            
            if not file_attachment_index:
                return data
            
            form = FormData()
            form.add_json('payload_json', data)
            
            for file_attachment_index, (name, io, description) in enumerate(
                details for is_attachment, details in value if not is_attachment
            ):
                form.add_field(
                    f'files[{file_attachment_index}]', io, file_name = name, content_type = 'application/octet-stream'
                )
            
            return form
    
    else:
        def serializer_putter(data, required, value):
            if value is None:
                return data
            
            file_attachments = [details for is_attachment, details in value if not is_attachment]
            if not file_attachments:
                return data
            
            form = FormData()
            form.add_json('payload_json', data)
            
            if len(file_attachments) == 1:
                name, io, description = file_attachments
                form.add_field('file', io, file_name = name, content_type = 'application/octet-stream')
            else:
                for file_attachment_index, (name, io, description) in enumerate(file_attachments):
                    form.add_field(
                        f'file{file_attachment_index}s', io, file_name = name, content_type = 'application/octet-stream'
                    )
            
            return form
    
    
    # Sorting
    
    sort_priority = 9999



__all__ = ()

from ...core import (
    APPLICATION_COMMANDS, AUTO_MODERATION_RULES, CHANNELS, EMOJIS, INTEGRATIONS, ROLES, SCHEDULED_EVENTS,
    SOUNDBOARD_SOUNDS, STAGES, STICKERS, USERS
)
from ...invite import Invite


def target_converter_guild(entry):
    return entry.guild


def target_converter_channel(entry):
    target_id = entry.target_id
    if target_id:
        return CHANNELS.get(target_id, None)


def target_converter_user(entry):
    target_id = entry.target_id
    if target_id:
        return USERS.get(target_id, None)


def target_converter_role(entry):
    target_id = entry.target_id
    if target_id:
        return ROLES.get(target_id, None)


def target_converter_invite(entry):
    # every other data is at # change
    change = entry.get_change('code')
    if change is None:
        code = None
    elif change.has_after():
        code = change.after
    elif change.has_before():
        code = change.before
    else:
        code = None
    
    if code is None:
        return None
    
    return Invite.precreate(code, guild = entry.guild)


def target_converter_webhook(entry):
    target_id = entry.target_id
    if target_id:
        return USERS.get(target_id, None)


def target_converter_emoji(entry):
    target_id = entry.target_id
    if target_id:
        return EMOJIS.get(target_id, None)


def target_converter_integration(entry):
    target_id = entry.target_id
    if target_id:
        return INTEGRATIONS.get(target_id, None)


def target_converter_stage(entry):
    target_id = entry.target_id
    if target_id:
        return STAGES.get(target_id, None)


def target_converter_sticker(entry):
    target_id = entry.target_id
    if target_id:
        return STICKERS.get(target_id, None)


def target_converter_scheduled_event(entry):
    target_id = entry.target_id
    if target_id:
        return SCHEDULED_EVENTS.get(target_id, None)


def target_converter_application_command(entry):
    target_id = entry.target_id
    if target_id:
        return APPLICATION_COMMANDS.get(target_id, None)


def target_converter_auto_moderation_rule(entry):
    target_id = entry.target_id
    if target_id:
        return AUTO_MODERATION_RULES.get(target_id, None)


def target_converter_soundboard_sound(entry):
    target_id = entry.target_id
    if target_id:
        return SOUNDBOARD_SOUNDS.get(target_id, None)



from .command_base import *
from .command_base_application_command import *
from .command_base_custom_id import *
from .component_command import *
from .embedded_activity_launch_command import *
from .context_command import *
from .form_submit_command import *
from .slash_command import *

from .helpers import *


__all__ = (
    *command_base.__all__,
    *command_base_application_command.__all__,
    *command_base_custom_id.__all__,
    *component_command.__all__,
    *embedded_activity_launch_command.__all__,
    *context_command.__all__,
    *form_submit_command.__all__,
    *slash_command.__all__,
    
    *helpers.__all__,
)



__all__ = ('PlaceHolderFunctional',)

from scarletio import copy_docs

from .base import PlaceHolderBase


class PlaceHolderFunctional(PlaceHolderBase):
    __type_doc__ = (
    """
    Slot place holder returning a default value.
    
    Might be used to avoid `__getattr__` definitions.
    
    Attributes
    ----------
    attribute_name : `None | str`
        The name of the place held attribute.
    
    default_function : `callable`
        A function to create the default value.
       
    docs : `None | str`
        Documentation of the place held attribute.
    
    type_name : `None | str`
        The name of the type the we place held at.
    """)
    
    __slots__ = ('default_function',)
    
    
    def __new__(cls, default_function, docs = None):
        """
        Creates a new new slot place holder.
        
        Parameters
        ----------
        default_function : `callable`
            A function to create the default value.
        
        docs : `None`, `str` = `None`, Optional
            Documentation of the place held attribute.
        """
        self = PlaceHolderBase.__new__(cls, docs)
        self.default_function = default_function
        return self
    
    
    @copy_docs(PlaceHolderBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        if not PlaceHolderBase._is_equal_same_type(self, other):
            return False
        
        # default_function
        if self.default_function != other.default_function:
            return False
        
        return True
    
    
    @copy_docs(PlaceHolderBase._put_repr_parts_into)
    def _put_repr_parts_into(self, repr_parts, field_added):
        # default_function
        if field_added:
            repr_parts.append(',')
        else:
            field_added = True
        
        repr_parts.append(' default_function = ')
        repr_parts.append(repr(self.default_function))
        
        return field_added
    
    
    @copy_docs(PlaceHolderBase.__hash__)
    def __hash__(self):
        hash_value = PlaceHolderBase.__hash__(self)
        
        # default_function
        default_function = self.default_function
        try:
            default_function_hash = hash(default_function)
        except TypeError:
            default_function_hash = object.__hash__(default_function)
        hash_value ^= default_function_hash
        
        return hash_value
    
    
    @copy_docs(PlaceHolderBase.__get__)
    def __get__(self, instance, type_):
        if instance is None:
            return self
        
        return self.default_function()



__all__ = ('GuildWidget',)

from ...bases import DiscordEntity
from ...http import urls as module_urls

from ..guild import Guild

from .fields import (
    parse_approximate_online_count, parse_channels, parse_id, parse_invite_url, parse_name, parse_users,
    put_approximate_online_count, put_channels, put_id, put_invite_url, put_name,
    put_users, validate_approximate_online_count, validate_channels, validate_id, validate_invite_url,
    validate_name, validate_users
)


class GuildWidget(DiscordEntity):
    """
    Represents a ``Guild``'s widget.
    
    Attributes
    ----------
    approximate_online_count : `int`
        Estimated count of the online users in the respective guild.
    channels : `None`, `tuple` of ``GuildWidgetChannel``
        Voice channels received with the guild widget.
    id : `int`
        The unique identifier number of the guild widget's guild.
    invite_url : `None`, `str`
        The guild widget's invite url if applicable.
    name : `str`
        The name of the guild widget's guild.
    users : `None`, `tuple` of ``GuildWidgetUser``
        Online users received with the guild widget.
    """
    __slots__ = ('approximate_online_count', 'channels', 'invite_url', 'name', 'users')
    
    def __new__(
        cls,
        *,
        approximate_online_count = ...,
        channels = ...,
        guild_id = ...,
        invite_url = ...,
        name = ...,
        users = ...,
    ):
        """
        Creates a new guild widget from the given fields.
        
        Parameters
        ----------
        approximate_online_count : `int`, Optional (Keyword only)
            Estimated count of the online users in the respective guild.
        channels : `None`, `iterable` of ``GuildWidgetChannel``, Optional (Keyword only)
            Voice channels received with the guild widget.
        id : `int`, Optional (Keyword only)
            The unique identifier number of the guild widget's guild.
        invite_url : `None`, `str`, Optional (Keyword only)
            The guild widget's invite url if applicable.
        name : `str`, Optional (Keyword only)
            The name of the guild widget's guild.
        users : `None`, `iterable` of ``GuildWidgetUser``, Optional (Keyword only)
            Online users received with the guild widget.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # approximate_online_count
        if approximate_online_count is ...:
            approximate_online_count = 0
        else:
            approximate_online_count = validate_approximate_online_count(approximate_online_count)
        
        # channels
        if channels is ...:
            channels = None
        else:
            channels = validate_channels(channels)
        
        # guild_id
        if guild_id is ...:
            guild_id = 0
        else:
            guild_id = validate_id(guild_id)
        
        # invite_url
        if invite_url is ...:
            invite_url = None
        else:
            invite_url = validate_invite_url(invite_url)
        
        # name
        if name is ...:
            name = ''
        else:
            name = validate_name(name)
        
        # users
        if users is ...:
            users = None
        else:
            users = validate_users(users)
        
        # Construct
        self = object.__new__(cls)
        self.approximate_online_count = approximate_online_count
        self.channels = channels
        self.id = guild_id
        self.invite_url = invite_url
        self.name = name
        self.users = users
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new guild widget from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            The requested guild widget data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.approximate_online_count = parse_approximate_online_count(data)
        self.channels = parse_channels(data)
        self.id = parse_id(data)
        self.invite_url = parse_invite_url(data)
        self.name = parse_name(data)
        self.users = parse_users(data)
        return self
    
    
    def to_data(self, * , defaults = False):
        """
        Serializes the guild widget.
        
        Parameters
        ----------
        defaults : `bool`
            Whether fields of their default value should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        data = {}
        put_approximate_online_count(self.approximate_online_count, data, defaults)
        put_channels(self.channels, data, defaults)
        put_id(self.id, data, defaults)
        put_invite_url(self.invite_url, data, defaults)
        put_name(self.name, data, defaults)
        put_users(self.users, data, defaults)
        return data
    
    
    def __repr__(self):
        """Returns the representation of the guild widget."""
        return f'<{self.__class__.__name__} id = {self.id!r}, name = {self.name!r}>'
    
    
    def __hash__(self):
        """Returns the has value of the guild widget."""
        hash_value = 0
        
        # approximate_online_count
        hash_value ^= self.approximate_online_count
        
        # channels
        channels = self.channels
        if (channels is not None):
            hash_value ^= len(channels) << 8
            
            for channel in channels:
                hash_value ^= hash(channel)
        
        # id
        hash_value ^= self.id
        
        # invite_url
        invite_url = self.invite_url
        if (invite_url is not None):
            hash_value ^= hash(invite_url)
        
        # name
        hash_value ^= hash(self.name)
        
        # users
        users = self.users
        if (users is not None):
            hash_value ^= len(users) << 12
            
            for user in users:
                hash_value ^= hash(user)
        
        return hash_value
    
    def __eq__(self, other):
        """Returns whether the two guild widget are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal_same_type(other)
    
    
    def __ne__(self, other):
        """Returns whether the two guild widget are not equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return not self._is_equal_same_type(other)
    
    
    def _is_equal_same_type(self, other):
        """
        Returns whether the two types are equal.
        
        Helper method for ``.__eq__``
        
        Parameters
        ----------
        other : `instance<type<self>>`
            The other instance. Must be from the same type.
        
        Returns
        -------
        is_equal : `bool`
        """
        # approximate_online_count
        if self.approximate_online_count != other.approximate_online_count:
            return False
        
        # channels
        if self.channels != other.channels:
            return False
        
        # id
        if self.id != other.id:
            return False
        
        # invite_url
        if self.invite_url != other.invite_url:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        # status
        if self.users != other.users:
            return False
        
        return True
    
    
    def copy(self):
        """
        Copies the guild widget.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.approximate_online_count = self.approximate_online_count
        
        channels = self.channels
        if (channels is not None):
            channels = (*(channel.copy() for channel in channels),)
        new.channels = channels
        
        new.id = self.id
        new.invite_url = self.invite_url
        new.name = self.name
        
        users = self.users
        if (users is not None):
            users = (*(user.copy() for user in users),)
        new.users = users
        
        return new
    
    
    def copy_with(
        self,
        *,
        approximate_online_count = ...,
        channels = ...,
        guild_id = ...,
        invite_url = ...,
        name = ...,
        users = ...,
    ):
        """
        Copies the guild widget from the given parameters.
        
        Parameters
        ----------
        approximate_online_count : `int`, Optional (Keyword only)
            Estimated count of the online users in the respective guild.
        channels : `None`, `iterable` of ``GuildWidgetChannel``, Optional (Keyword only)
            Voice channels received with the guild widget.
        id : `int`, Optional (Keyword only)
            The unique identifier number of the guild widget's guild.
        invite_url : `None`, `str`, Optional (Keyword only)
            The guild widget's invite url if applicable.
        name : `str`, Optional (Keyword only)
            The name of the guild widget's guild.
        users : `None`, `iterable` of ``GuildWidgetUser``, Optional (Keyword only)
            Online users received with the guild widget.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # approximate_online_count
        if approximate_online_count is ...:
            approximate_online_count = self.approximate_online_count
        else:
            approximate_online_count = validate_approximate_online_count(approximate_online_count)
        
        # channels
        if channels is ...:
            channels = self.channels
            if (channels is not None):
                channels = (*(channel.copy() for channel in channels),)
        else:
            channels = validate_channels(channels)
        
        # guild_id
        if guild_id is ...:
            guild_id = self.id
        else:
            guild_id = validate_id(guild_id)
        
        # invite_url
        if invite_url is ...:
            invite_url = self.invite_url
        else:
            invite_url = validate_invite_url(invite_url)
        
        # name
        if name is ...:
            name = self.name
        else:
            name = validate_name(name)
        
        # users
        if users is ...:
            users = self.users
            if (users is not None):
                users = (*(user.copy() for user in users),)
        else:
            users = validate_users(users)
        
        # Construct
        new = object.__new__(type(self))
        new.approximate_online_count = approximate_online_count
        new.channels = channels
        new.id = guild_id
        new.invite_url = invite_url
        new.name = name
        new.users = users
        return new
    
    
    json_url = property(module_urls.guild_widget_json_url)
    
    
    @property
    def guild(self):
        """
        Returns the parent guild of the widget. If it is not in cache creates a new one.
        
        Returns
        -------
        guild : ``Guild``
        """
        return Guild.precreate(self.id, name = self.name)
    
    
    def iter_channels(self):
        """
        Iterates over the widget channels.
        
        This method is an iterable generator.
        
        Yields
        ------
        channel : ``GuildWidgetChannel``
        """
        channels = self.channels
        if channels is not None:
            yield from channels
    
    
    def iter_users(self):
        """
        Iterates over the widget users.
        
        This method is an iterable generator.
        
        Yields
        ------
        user : ``GuildWidgetUser``
        """
        users = self.users
        if users is not None:
            yield from users



__all__ = ()

from ...emoji import Emoji, create_partial_emoji_data, create_partial_emoji_from_data
from ...field_parsers import (
    entity_id_parser_factory, float_parser_factory, nullable_functional_parser_factory, preinstanced_parser_factory
)
from ...field_putters import (
    entity_id_optional_putter_factory, entity_id_putter_factory, float_optional_putter_factory,
    nullable_functional_optional_putter_factory, preinstanced_optional_putter_factory
)
from ...field_validators import (
    entity_id_validator_factory, float_conditional_validator_factory, nullable_entity_validator_factory,
    preinstanced_validator_factory
)
from ...soundboard import SoundboardSound
from ...user import ClientUserBase

from ..channel import Channel

from .preinstanced import VoiceChannelEffectAnimationType


# animation_id

parse_animation_id = entity_id_parser_factory('animation_id')
put_animation_id = entity_id_optional_putter_factory('animation_id')
validate_animation_id = entity_id_validator_factory('animation_id')


# animation_type

parse_animation_type = preinstanced_parser_factory(
    'animation_type', VoiceChannelEffectAnimationType, VoiceChannelEffectAnimationType.premium
)
put_animation_type = preinstanced_optional_putter_factory(
    'animation_type', VoiceChannelEffectAnimationType.premium
)
validate_animation_type = preinstanced_validator_factory('animation_type', VoiceChannelEffectAnimationType)


# channel_id

parse_channel_id = entity_id_parser_factory('channel_id')
put_channel_id = entity_id_putter_factory('channel_id')
validate_channel_id = entity_id_validator_factory('channel_id', Channel)


# emoji

parse_emoji = nullable_functional_parser_factory('emoji', create_partial_emoji_from_data)
put_emoji = nullable_functional_optional_putter_factory('emoji', create_partial_emoji_data)
validate_emoji = nullable_entity_validator_factory('emoji', Emoji)


# guild_id

parse_guild_id = entity_id_parser_factory('guild_id')
put_guild_id = entity_id_putter_factory('guild_id')
validate_guild_id = entity_id_validator_factory('guild_id', NotImplemented, include = 'Guild')


# sound_id

parse_sound_id = entity_id_parser_factory('sound_id')
put_sound_id = entity_id_optional_putter_factory('sound_id')
validate_sound_id = entity_id_validator_factory('sound_id', SoundboardSound)


# sound_volume

parse_sound_volume = float_parser_factory('sound_volume', 1.0)
put_sound_volume = float_optional_putter_factory('sound_volume', 1.0)
validate_sound_volume = float_conditional_validator_factory(
    'sound_volume',
    1.0,
    lambda sound_volume : sound_volume >= 0.0 and sound_volume <= 1.0,
    '>= 0.0 and <= 1.0',
)


# user_id

parse_user_id = entity_id_parser_factory('user_id')
put_user_id = entity_id_putter_factory('user_id')
validate_user_id = entity_id_validator_factory('user_id', ClientUserBase)



__all__ = ()

import sys
from os import listdir as list_directory, system as execute
from os.path import (
    abspath as absolute_path, getctime as get_creation_time, isdir as is_directory, isfile as is_file,
    join as join_paths, splitext as split_extension
)
from shlex import quote

from scarletio import get_short_executable

from ...core import LIBRARY_CALLED_DIRECTLY, register
from ...core.helpers import render_main_call_into

# Conditional imports
if sys.platform == 'linux':
    from os import geteuid as get_effective_user_id
else:
    get_effective_user_id = None


PYTHON_EXTENSIONS = frozenset(('.py', '.pyd', '.pyc', '.so'))


def get_profile_names():
    """
    Gets the profiling results' names.
    
    Returns
    -------
    names : `list<str>`
    """
    directory_path = absolute_path('.profiles')
    
    if not is_directory(directory_path):
        return []
    
    results = []
    for name in list_directory(directory_path):
        name_no_extension, extension = split_extension(name)
        if extension != '.prof':
            continue
        
        file_path = join_paths(directory_path, name)
        if not is_file(file_path):
            continue
        
        key = get_creation_time(file_path)
        results.append((key, name_no_extension))
    
    results.sort(reverse = True)
    return [item[1] for item in results]


def get_profile_path(name):
    """
    Get the profile's path for the given name.
    
    Parameters
    ----------
    name : `str`
        Profile name.
    
    Returns
    -------
    path : `None | str`
    """
    directory_path = absolute_path('.profiles')
    path = join_paths(directory_path, name + '.prof')
    if is_file(path):
        return path
    
    return None


def build_profile_listing(names):
    """
    Builds profile listing.
    
    Parameters
    ----------
    names : `list<str>`
        The names of profiles.
    
    Returns
    -------
    output : `str`
    """
    output_parts = []
    for name in names:
        output_parts.append('- ')
        output_parts.append(name)
        output_parts.append('\n')
    
    return ''.join(output_parts)


def has_package(name):
    """
    Returns whether the package with the given name exists.
    
    Parameters
    ----------
    name : `str`
        The name of the package to look up.
    
    Returns
    -------
    has_package : `bool`
    """
    for path in sys.path:
        package_path = join_paths(path, name)
        if is_directory(package_path):
            return True
        
        for extension in PYTHON_EXTENSIONS:
            if is_file(package_path + extension):
                return True
    
    return False


def build_profile_path_not_found(name):
    """
    Builds a profile path not found message.
    
    Parameters
    ----------
    name : `str`
        The profile's name.
    
    Returns
    -------
    output : `str`
    """
    output_parts = []
    output_parts.append('Profile file with name: ')
    output_parts.append(repr(name))
    output_parts.append(' not found.\nTry using "$ ')
    render_main_call_into(output_parts)
    output_parts.append(' profile list" to list the available profiling results.\n')
    return ''.join(output_parts)


def build_package_not_installed(name):
    """
    Builds a package not installed message.
    
    Parameters
    ----------
    name : `str`
        Package name.
    
    Returns
    -------
    output : `str`
    """
    output_parts = []
    output_parts.append('Required package: ')
    output_parts.append(repr(name))
    output_parts.append(' is not installed, cannot show profiling result.\n\nTo install it do:\n\n```\n$ ')
    
    if (get_effective_user_id is not None) and (get_effective_user_id() == 0):
        output_parts.append('sudo ')
    
    output_parts.append(quote(get_short_executable()))
    output_parts.append(' -m pip install ')
    output_parts.append(quote(name))
    output_parts.append('\n```\n')
    return ''.join(output_parts)


PROFILING_CATEGORY = register(
    None,
    available = (not LIBRARY_CALLED_DIRECTLY),
    name = 'profiling',
)


@register(
    into = PROFILING_CATEGORY,
)
def list_():
    """
    Lists the available profiling results.
    """
    names = get_profile_names()
    if not names:
        return 'No profiling results available\n'
    
    return build_profile_listing(names)


@register(
    into = PROFILING_CATEGORY,
)
def show(
    name : str = 'latest',
):
    """
    Shows the given or the latest profiling result.
    """
    path = get_profile_path(name)
    if name is None:
        return build_profile_path_not_found(name)
    
    if not has_package('snakeviz'):
        return build_package_not_installed('snakeviz')
    
    execute(f'{quote(get_short_executable())!s} -m snakeviz {quote(path)!s}')



__all__ = ('CONVERSION_COMPONENTS',)

from ....builder.constants import CONVERSION_KIND_FIELD
from ....builder.conversion import Conversion
from ....component import Component, ComponentType, create_row


def _is_component_listing(value):
    """
    Yields the outcome if the input `value` is a component listing. `value` must be a non-empty list.
    
    This function is a generator.    
    
    Parameters
    ----------
    value : `(list | tuple)<object>`
        Value to identify.
    
    Yields
    ------
    components : `list<Component>`
    """
    built = None
    
    for element in value:
        if isinstance(element, Component):
            if not element.type.layout_flags.top_level:
                element = create_row(element)
        
        elif isinstance(element, list) or isinstance(element, tuple):
            if not all(isinstance(nested_element, Component) for nested_element in element):
                return
            
            element = create_row(*element)
        
        else:
            return
        
        if built is None:
            built = []
        built.append(element)
        continue
    
    yield built
    return



class CONVERSION_COMPONENTS(Conversion):
    # Generic
    
    name = 'components'
    name_aliases = None
    expected_types_messages = (
        f'``None`, ``{Component.__name__}``, (`list`, `tuple`) of (``{Component.__name__}``, '
        f'(`list`, `tuple`) of ``{Component.__name__}``)'
    )
    kind = CONVERSION_KIND_FIELD
    output_conversion = None
    
    # Setting
    
    set_merger = None
    
    set_type = Component
    
    
    def set_type_processor(value):
        if not value.type.layout_flags.top_level:
            value = create_row(value)
        
        return [value]
    
    
    set_listing_identifier = _is_component_listing
    set_identifier = None
    
    
    def set_validator(value):
        # None
        if value is None:
            yield None
            return
        
        # Component
        if isinstance(value, Component):
            if not value.type.layout_flags.top_level:
                value = create_row(value)
            
            yield [value]
            return
        
        # (list | tuple)<length = 0> | (list | tuple)<Component | (list | tuple)<Component>>
        if isinstance(value, list) or isinstance(value, tuple):
            if not value:
                yield None
                return
            
            yield from _is_component_listing(value)
            return
        
        # No other cases
        return
    
    
    # Reading
    
    get_default = None
    get_processor = None
    
    
    # Serialization
    
    serializer_key = 'components'
    
    def serializer_optional(value):
        if value is not None:
            yield [element.to_data() for element in value]
    
    
    def serializer_required(value):
        if value is None:
            return []
        
        return [element.to_data() for element in value]
    
    
    # Sorting
    
    sort_priority = 1400



__all__ = ('AuditLogIterator', )

from warnings import warn

from scarletio import RichAttributeErrorBaseType, Task, shield

from ...core import KOKORO
from ...utils import now_as_id

from ..audit_log import AuditLog

from .fields import validate_entry_type, validate_guild_id, validate_user_id


class AuditLogIterator(RichAttributeErrorBaseType):
    """
    An asynchronous iterator over a guild's audit logs.
    
    Attributes
    ----------
    _audit_log_entry_index : `int`
        The index of the entry in an audit to return next.
    _audit_log_index : `int`
        The index of the audit log to return entry from.
    _data : `dict<str, object>`
        Data to make request with.
    _load_one_task : `None | Task<bool>`
        Synchronised task among consumers to request a new audit log.
    audit_logs : `list` of ``AuditLog``
        The requested audit logs.
    client : ``Client``
        Client to request with.
    guild_id : `int`
        The respective guild's identifier.
    """
    __slots__ = (
        '_audit_log_entry_index', '_audit_log_index', '_data', '_load_one_task', 'audit_logs', 'client', 'guild_id'
    )
    
    def __new__(cls, client, guild_id, *, entry_type = ..., user_id = ..., user = ...):
        """
        Creates an audit log iterator with the given parameters.
        
        Parameters
        ----------
        client : ``Client``
            The client, who will execute the api requests.
        guild_id : ``Guild``, `int`
            The guild what's audit logs will be requested.
        entry_type : ``AuditLogEntryType``, `int`, `None`, Optional
            Whether the audit logs should be filtered only on the given event.
        user_id : `int`, `None`, ``ClientUserBase``, Optional
            Whether the audit logs should be filtered only to those, which were created by the given user.
        """
        if user is not ...:
            warn(
                f'`{cls.__name__}.__new__`\'s `user` parameter is deprecated and will be removed in 2024 December. '
                f'Please use `user_id` instead.'
            )
            user_id = user
        
        # guild_id
        guild_id = validate_guild_id(guild_id)
        
        # Build data
        data = {
            'limit': 100,
            'before': now_as_id(),
        }
        
        # entry_type
        if (entry_type is not ...):
            data['action_type'] = validate_entry_type(entry_type).value
        
        # user_id
        if (user_id is not ...):
            user_id = validate_user_id(user_id)
            if user_id:
                data['user_id'] = str(user_id)
        
        
        self = object.__new__(cls)
        self._audit_log_entry_index = 0
        self._audit_log_index = 0
        self._data = data
        self._load_one_task = None
        self.audit_logs = []
        self.client = client
        self.guild_id = guild_id
        
        return self
    
    
    async def _load_one_inner(self):
        """
        Loads a single audit log.
        
        This method is a coroutine.
        """
        client = self.client
        data = self._data
        
        audit_log_data = await client.api.audit_log_get_chunk(self.guild_id, data)
        audit_log = AuditLog.from_data(audit_log_data, self.guild_id)
        self.audit_logs.append(audit_log)
        
        audit_log_length = len(audit_log)
        if audit_log_length < 100:
            return False
        
        data['before'] = audit_log[-1].id
        return True
    
    
    async def _load_one_outer(self):
        """
        Loads a single audit log. Does synchronised ``._load_one_inner`` calls.
        
        This method is a coroutine.
        """
        load_one_task = self._load_one_task
        if load_one_task is None or load_one_task.is_done():
            load_one_task = Task(KOKORO, self._load_one_inner())
            self._load_one_task = load_one_task
        
        try:
            return await shield(load_one_task, KOKORO)
        finally:
            if load_one_task.is_done():
                self._load_one_task = None
    
    
    async def load_all(self):
        """
        Loads all not yet loaded audit logs of the audit log iterator's guild.
        
        This method is a coroutine.
        """
        while (await self._load_one_outer()):
            pass
    
    
    def transform(self):
        """
        Converts the audit log iterator to an audit log object.
        
        Returns
        -------
        audit_log : ``AuditLog``
        """
        return AuditLog.from_many(self.audit_logs)
    
    
    def __aiter__(self):
        """Returns self and resets the `.index`."""
        return self
    
    
    async def __anext__(self):
        """
        Yields the next entry of the audit log iterator.
        
        This method is a coroutine.
        """
        audit_logs = self.audit_logs
        
        # Use a while True loop, because if over 100 parallel iterations are done it may mean that some consumers will
        # not be satisfied from 1 pull.
        while True:
            audit_logs_length = len(audit_logs)
            audit_log_index = self._audit_log_index
            
            if audit_logs_length > audit_log_index:
                audit_log = audit_logs[audit_log_index]
                
                audit_log_entries_length = len(audit_log)
                audit_log_entry_index = self._audit_log_entry_index
                if audit_log_entry_index < audit_log_entries_length:
                    entry = audit_log[audit_log_entry_index]
                    self._audit_log_entry_index += 1
                    return entry
                
                if audit_log_entries_length < 100:
                    raise StopAsyncIteration
                
                self._audit_log_index += 1
                self._audit_log_entry_index = 0
            
            await self._load_one_outer()
    
    
    def __repr__(self):
        """Returns the audit log iterator's representation."""
        repr_parts = ['<', type(self).__name__]
        
        # client
        repr_parts.append(' client = ')
        repr_parts.append(repr(self.client))
        
        # guild_id
        repr_parts.append(', guild_id = ')
        repr_parts.append(repr(self.guild_id))
        
        repr_parts.append('>')
        return ''.join(repr_parts)



__all__ = ()

from scarletio import RichAttributeErrorBaseType, WeakReferer


def get_self_reference_of(instance):
    """
    Gets self reference of the given instance.
    
    Parameters
    ----------
    instance : `None | SelfReferenceInterface | object`
        Instance to get self reference of.
    
    Returns
    -------
    self_reference: `None | WeakReferer`
    """
    if instance is None:
        return None
    
    if isinstance(instance, SelfReferenceInterface):
        return instance.get_self_reference()
    
    return None


class SelfReferenceInterface(RichAttributeErrorBaseType):
    """
    Base class for types supporting self-reference getting.
    """
    __slots__ = ()
    
    def get_self_reference(self):
        """
        Gets a weak reference to the instance.
        
        Returns
        -------
        self_reference : `WeakReferer<self>`
        """
        self_reference = self._self_reference
        if self_reference is None:
            try:
                self_reference = WeakReferer(self)
            except TypeError:
                self_reference = None
            else:
                self._self_reference = self_reference
        
        return self_reference
    
    
    @property
    def _self_reference(self):
        """
        The registered self-reference.
        
        Returns
        -------
        self_reference : `None | WeakReferer`
        """
        return None
    
    
    @_self_reference.setter
    def _self_reference(self, value):
        pass



__all__ = ()

from ...scheduled_event import PrivacyLevel
from ...stage.stage.fields import validate_privacy_level, validate_topic

from ..audit_log_entry_change_conversion import AuditLogEntryChangeConversion, AuditLogEntryChangeConversionGroup
from ..conversion_helpers.converters import value_deserializer_description, value_serializer_description


# ---- privacy_level ----

PRIVACY_LEVEL_CONVERSION = AuditLogEntryChangeConversion(
    ('privacy_level',),
    'privacy_level',
    value_validator = validate_privacy_level,
)


@PRIVACY_LEVEL_CONVERSION.set_value_deserializer
def privacy_level_value_deserializer(value):
    return PrivacyLevel(value)


@PRIVACY_LEVEL_CONVERSION.set_value_serializer
def privacy_level_value_serializer(value):
    return value.value


# ---- topic ----

TOPIC_CONVERSION = AuditLogEntryChangeConversion(
    ('topic',),
    'topic',
    value_deserializer = value_deserializer_description,
    value_serializer = value_serializer_description,
    value_validator = validate_topic,
)


# ---- Construct ----

STAGE_CONVERSIONS = AuditLogEntryChangeConversionGroup(
    PRIVACY_LEVEL_CONVERSION,
    TOPIC_CONVERSION,
)
